from mcstasscript.helper.formatting import bcolors
from mcstasscript.helper.formatting import is_legal_parameter
from mcstasscript.helper.exceptions import McStasError
from mcstasscript.helper.name_inspector import find_python_variable_name
from mcstasscript.helper.search_statement import SearchStatement, SearchStatementList

from libpyvinyl.Parameters.Parameter import Parameter


def provide_parameter(*args, **kwargs):
    """Makes a libpyvinyl parameter object

    Examples
    --------

    Creates a parameter with name wavelength and associated comment
    A = provide_parameter("wavelength", comment="wavelength in [AA]")

    Creates a parameter with name A3 and default value
    A = provide_parameter("A3", value=30, comment="A3 angle in [deg]")

    Creates a parameter with type string and name sample_name
    A = provide_parameter("string", "sample_name")

    Parameters
    ----------
    If giving a type:
    Positional argument 1: type: str
        Type of the parameter, double, int or string
    Positional argument 2: name: str
        Name of input parameter

    If not giving type
    Positional argument 1: name : str
        Name of input parameter

    Keyword arguments
        value : float, int or str
            sets default value of parameter
        unit : str
            string that describes the unit
        comment : str
            sets comment displayed next to declaration
        options : list or value
            list or value of allowed values for this parameter
    """
    if len(args) == 0:
        # Check all required keyword arguments present
        if "name" not in kwargs:
            try:
                name = find_python_variable_name(error_text="", n_levels=3)
                kwargs["name"] = name
            except:
                raise RuntimeError("Need to provide name, either as first argument"
                                   + ", keyword argument or python variable.")

        provided_name = kwargs["name"]

        provided_type = ""
        if "type" in kwargs:
            provided_type = kwargs["type"]

    elif len(args) == 1:
        if "name" in kwargs:
            raise RuntimeError("Only specify name with argument or keyword argument")
        provided_name = args[0]

        # Assume default type if not given
        provided_type = ""
        if "type" in kwargs:
            provided_type = kwargs["type"]

    elif len(args) == 2:
        if "type" in kwargs:
            raise RuntimeError("Only specify type with argument or keyword argument")
        provided_type = args[0]

        if "name" in kwargs:
            raise RuntimeError("Only specify name with argument or keyword argument")
        provided_name = args[1]

    else:
        raise RuntimeError("Too many arguments given to parameter")

    if not is_legal_parameter(provided_name):
        raise NameError("The given parameter name: \""
                        + provided_name
                        + "\" is not a legal c variable name, "
                        + " and cannot be used in McStas.")

    allowed_types = {"", "double", "int", "string"}
    if provided_type not in allowed_types:
        raise RuntimeError("Tried to create parameter of type \""
                           + str(provided_type)
                           + "\" which is not among the allowed types "
                           + str(allowed_types) + ".")

    comment = ""
    if "comment" in kwargs:
        if not isinstance(comment, str):
            raise RuntimeError("Tried to create a parameter with a "
                               + "comment that was not a string.")
        comment = kwargs["comment"]

    unit = ""
    if "unit" in kwargs:
        if not isinstance(unit, str):
            raise RuntimeError("Unit has to be a string")
        unit = kwargs["unit"]

    parameter = Parameter(name=provided_name, unit=unit, comment=comment)
    parameter.type = provided_type

    if "options" in kwargs:
        parameter.add_option(kwargs["options"], True)

    if "value" in kwargs:
        parameter.value = kwargs["value"]

    return parameter


def write_parameter(fo, parameter, stop_character):
    """
    Writes a parameter object to McStas define section

    Parameters
    ----------

    fo : file object
        Open file object to write parameter string to

    parameter : Parameter
        Parameter object to be written

    stop_character : str
        Character inserted after parameter, usually comma and space
    """
    if not isinstance(stop_character, str):
        raise RuntimeError("stop_character in write_parameter should be "
                           + "a string.")

    if parameter.type is None:
        # This can happen if parameter given by libpyvinyl directly, infer type
        if parameter.value is None:
            raise RuntimeError("Need to set parameter '" + parameter.name
                               + "' before writing instrument.")

        if isinstance(parameter.value, (float, int)):
            parameter.type = "double"
        elif isinstance(parameter.value, str):
            parameter.type = "string"
        else:
            raise RuntimeError("Parameter '" + parameter.name + "' has value "
                               + "of type not recognized by McStasScript.")

    if not parameter.type == "":
        fo.write("%s %s" % (parameter.type, parameter.name))
    else:
        fo.write(parameter.name)

    if parameter.value is not None:
        if isinstance(parameter.value, int):
            fo.write(" = %d" % parameter.value)
        elif isinstance(parameter.value, float):
            fo.write(" = %G" % parameter.value)
        else:
            fo.write(" = %s" % str(parameter.value))
    fo.write(stop_character)

    if parameter.comment is None or parameter.comment == "":
        c_comment = ""
    else:
        c_comment = "// " + parameter.comment

    fo.write(c_comment)
    fo.write("\n")


class DeclareVariable:
    """
    Class describing a declared variable in McStas instrument

    McStas parameters are declared in declare section with c syntax.
    This class is initialized with type, name.  Using keyword
    arguments, the variable can become an array and have its initial
    value set.
    Can also be used as a user variable in uservars (McStas 3.0)

    Attributes
    ----------
    type : str
        McStas type to declare: Double, Int, String

    name : str
        Name of variable

    value : any
        Initial value of variable, converted to string

    comment : str
        Comment displayed next to the declaration, could contain units

    vector : int
        0 if a single value is given, otherwise contains the length

    Methods
    -------
    write_line(fo)
        Writes a line to text file fo declaring the parameter in c
    """
    def __init__(self, type, name=None, **kwargs):
        """
        Initializing mcstas declare variable or user variable object

        Examples
        --------

        Creates a variable with name A3 and default value
        A = DeclareVariable("double", "A3", value=30)

        Creates a variable with type integer and name A
        A = DeclareVariable("int")

        Creates a variable with type integer and name sample_number
        A = DeclareVariable("int", "sample_number")

        Creates an array variable called m_values
        A = DeclareVariable("double", "m_values", array=3,
                             value=[2, 2.5, 2])

        Parameters
        ----------
        type : str
            Type of the parameter, double, int or string

        name : str
            Name of input parameter

        Keyword arguments
            array : int
                length of array to be allocated, 0 if single value

            value : any
                sets initial value of parameter,
                can be a list with length matching array

            comment : str
                sets comment displayed next to declaration
        """

        self.type = type
        if not isinstance(self.type, str):
            raise RuntimeError("Given type of DeclareVariable should be a "
                               + "string.")

        if name is None:
            error_text = ("When using automatic assignment of name, the call"
                          " need to assign it to a variable name")
            name = find_python_variable_name(error_text=error_text, n_levels=3)

        self.name = str(name)

        par_name = self.name
        if "*" in par_name[0]:
            # Remove any number of prefixed *, indicating variable is a pointer
            par_name = par_name.split("*")[-1]
        elif "&" in par_name[0]:
            # Remove the first & indicating the variable is an address
            par_name = par_name[1:]

        if not is_legal_parameter(par_name):
            raise NameError("The given parameter name: \""
                            + self.name
                            + "\" is not a legal c variable name, "
                            + " and cannot be used in McStas.")

        self.value = ""
        if "value" in kwargs:
            self.value = kwargs["value"]

        self.vector = 0
        if "array" in kwargs:
            self.vector = kwargs["array"]

        self.comment = ""
        if "comment" in kwargs:
            self.comment = " // " + kwargs["comment"]

    def write_line(self, fo):
        """
        Writes line declaring variable to file fo

        Parameters
        ----------
        fo : file object
            File the line will be written to
        """

        if self.value == "" and self.vector == 0:
            fo.write("%s %s;%s" % (self.type, self.name, self.comment))
        if self.value != "" and self.vector == 0:
            if self.type == "int":
                fo.write("%s %s = %d;%s" % (self.type, self.name,
                                            self.value, self.comment))
            else:
                try:
                    fo.write("%s %s = %G;%s" % (self.type, self.name,
                                                self.value, self.comment))
                except TypeError:
                    # Value could not be converted to float, write as string
                    fo.write("%s %s = %s;%s" % (self.type, self.name,
                                                self.value, self.comment))
        if self.value == "" and self.vector != 0:
            fo.write("%s %s[%d];%s" % (self.type, self.name,
                                       self.vector, self.comment))
        if self.value != "" and self.vector != 0:
            if isinstance(self.value, str):
                # value is a string
                string = self.value
                fo.write("%s %s[%d] = %s;" % (self.type, self.name,
                                              self.vector, string))
            else:
                # list of values
                fo.write("%s %s[%d] = {" % (self.type, self.name, self.vector))
                for i in range(0, len(self.value) - 1):
                    fo.write("%G," % self.value[i])
                fo.write("%G};%s" % (self.value[-1], self.comment))

    def __repr__(self):
        string = "Declare variable: '"
        string += str(self.name)
        string += "' of type "
        string += str(self.type)

        if self.value != "":
            string += " with value: "
            string += str(self.value)

        if self.vector != 0:
            string += ". Array with length "
            string += str(self.vector)

        return string


class Component:
    """
    A class describing a McStas component to be written to an instrument

    This class is used by the instrument class when setting up
    components as dynamic subclasses to this class.  Most information
    can be given on initialize using keyword arguments, but there are
    methods for setting the attributes describing the component. The
    class contains both methods to write the component to an instrument
    file and methods for printing to the python terminal for checking
    the information. The McStas_Instr class creates subclasses from
    this class that have attributes for all parameters for the given
    component. The component information is read directly from the
    component files in the McStas installation.  This class is frozen
    after __init__ so that no new attributes can be created, which
    allows direct feedback to the user if a parameter name is
    misspelled.

    Attributes
    ----------
    name : str
        Name of the component instance in McStas (must be unique)

    component_name : str
        Name of the component code to use, e.g. Arm, Guide_gravity, ...

    AT_data : list of 3 floats, default [0, 0, 0]
        Position data of the component

    AT_relative : str, default "ABSOLUTE"
        Name of former component to use as reference for position

    ROTATED_data : list of 3 floats, default [0, 0, 0]
        Rotation data of the component

    ROTATED_relative : str, default "ABSOLUTE"
        Name of former component to use as reference for position

    WHEN : str, default ""
        String with logical c expression x for when component is active

    EXTEND : str, default ""
        c code for McStas EXTEND section

    GROUP : str, default ""
        Name of group the component should belong to

    JUMP : str, default ""
        String describing use of JUMP, need to contain all after "JUMP"

    SPLIT : int, default 0 (disabled)
        Integer setting SPLIT, splitting the neutron before this component

    c_code_before : str, default ""
        C code inserted before the component

    c_code_after : str, default ""
        C code inserted after the component

    component_parameters : dict
        Parameters to be used with component in dictionary

    comment : str, default ""
        Comment inserted before the component as an explanation

    __isfrozen : bool
        If true no new attributes can be created, when false they can

    Methods
    -------
    set_AT(at_list, RELATIVE)
        Sets AT_data, can set AT_relative using keyword

    set_AT_RELATIVE(relative)
        Can set RELATIVE for position

    set_ROTATED(rotated_list, RELATIVE)
        Sets ROTATED_data, can set ROTATED_relative using keyword

    set_ROTATED_RELATIVE(relative)
        Can set RELATIVE for rotation

    set_RELATIVE(relative_name)
        Set both AT_relative and ROTATED_relative to relative_name

    set_parameters(dict_input)
        Adds dictionary entries to parameter dictionary

    set_WHEN(string)
        Sets WHEN string

    set_SPLIT(value)
        Sets SPLIT value, a value of 0 disables SPLIT

    set_GROUP(string)
        Sets GROUP name

    set_JUMP(string)
        Sets JUMP string

    append_EXTEND(string)
        Append string to EXTEND string

    set_c_code_before(string)
        Sets c code to be inserted before component

    set_c_code_before(string)
        Sets c code to be inserted after component

    set_comment(string)
        Sets comment for component

    write_component(fo)
        Writes component code to instrument file

    show_parameters()
        Prints current component state with all parameters

    print_long()
        Prints basic view of component code (not correct syntax)

    print_long_deprecated()
        Prints basic view of component code (obsolete)

    print_short(**kwargs)
        Prints short description, used in print_components

    set_keyword_input(**kwargs)
        Handle keyword arguments during initialize

    __setattr__(key, value)
        Overwriting __setattr__ to implement ability to freeze

    _freeze()
        Freeze the class so no new attributes can be defined

    _unfreeze()
        Unfreeze the class so new attributes can be defined again
    """

    __isfrozen = False  # When frozen, no new attributes allowed

    def __init__(self, instance_name, component_name, AT=None,
                 AT_RELATIVE=None, ROTATED=None, ROTATED_RELATIVE=None,
                 RELATIVE=None, WHEN=None, EXTEND=None, GROUP=None,
                 JUMP=None, SPLIT=None, comment=None, c_code_before=None,
                 c_code_after=None, save_parameters=False):
        """
        Initializes McStas component with specified name and component

        Parameters
        ----------
        instance_name : str
            name of the instance of the component

        component_name : str
            name of the component type e.g. Arm, Guide_gravity, ...

        keyword arguments:
            AT : list of 3 floats, default [0, 0, 0]
                Sets AT_data describing position of component

            AT_RELATIVE : str, default "ABSOLUTE"
                sets AT_relative, describing position reference

            ROTATED : list of 3 floats, default [0, 0, 0]
                Sets ROTATED_data, describing rotation of component

            ROTATED_RELATIVE : str, default "ABSOLUTE"
                Sets ROTATED_relative, sets reference for rotation

            RELATIVE : str
                Sets both AT_relative and ROTATED_relative

            WHEN : str, default ""
                Sets WHEN string, should contain logical c expression

            EXTEND : str, default ""
                Sets initial EXTEND string, should contain c code

            GROUP : str, default ""
                Sets name of group the component should belong to

            JUMP : str, default ""
                Sets JUMP str

            SPLIT : int, default 0 (disabled)
                Sets SPLIT value

            comment: str, default ""
                Sets comment string

            c_code_before: str, default ""
                Sets c code before component

            c_code_after: str, default ""
                Sets c code after component
        """

        # Allow addition of attributes in init
        self._unfreeze()

        self.name = instance_name
        self.component_name = component_name

        # initialize McStas information
        self.AT_data = [0, 0, 0]
        self.AT_relative = "ABSOLUTE"
        self.ROTATED_specified = False
        self.ROTATED_data = [0, 0, 0]
        self.ROTATED_relative = "ABSOLUTE"
        self.WHEN = ""
        self.EXTEND = ""
        self.GROUP = ""
        self.JUMP = ""
        self.SPLIT = 0
        self.comment = ""
        self.c_code_before = ""
        self.c_code_after = ""
        self.search_statement_list = SearchStatementList()
        self.save_parameters = save_parameters

        # references to component names
        self.AT_reference = None
        self.ROTATED_reference = None

        # If any keywords are set in kwargs, update these
        self.set_keyword_input(AT=AT, AT_RELATIVE=AT_RELATIVE, ROTATED=ROTATED,
                               ROTATED_RELATIVE=ROTATED_RELATIVE,
                               RELATIVE=RELATIVE, WHEN=WHEN, EXTEND=EXTEND,
                               GROUP=GROUP, JUMP=JUMP, SPLIT=SPLIT,
                               comment=comment, c_code_before=c_code_before,
                               c_code_after=c_code_after)

        """
        Could store an option for whether this component should be
        printed in instrument file or in a separate file which would
        then be included.
        """

        # Do not allow addition of attributes after init
        self._freeze()

    def set_keyword_input(self, AT=None, AT_RELATIVE=None, ROTATED=None,
                          ROTATED_RELATIVE=None, RELATIVE=None, WHEN=None,
                          EXTEND=None, GROUP=None, JUMP=None, SPLIT=None,
                          comment=None, c_code_before=None, c_code_after=None,
                          save_parameters=None):
        # Allow addition of attributes in init
        self._unfreeze()

        if AT is not None:
            self.set_AT(AT)

        if AT_RELATIVE is not None:
            self.set_AT_RELATIVE(AT_RELATIVE)

        self.ROTATED_specified = False
        if ROTATED is not None:
            self.set_ROTATED(ROTATED)

        if ROTATED_RELATIVE is not None:
            self.set_ROTATED_RELATIVE(ROTATED_RELATIVE)

        if RELATIVE is not None:
            self.set_RELATIVE(RELATIVE)

        if WHEN is not None:
            self.set_WHEN(WHEN)

        if EXTEND is not None:
            self.append_EXTEND(EXTEND)

        if GROUP is not None:
            self.set_GROUP(GROUP)

        if JUMP is not None:
            self.set_JUMP(JUMP)

        if SPLIT is not None:
            self.set_SPLIT(SPLIT)

        if comment is not None:
            self.set_comment(comment)

        if c_code_before is not None:
            self.set_c_code_before(c_code_before)

        if c_code_after is not None:
            self.set_c_code_after(c_code_after)

    def __setattr__(self, key, value):
        if self.__isfrozen and not hasattr(self, key):
            raise AttributeError("No parameter called '"
                                 + key
                                 + "' in component named "
                                 + self.name
                                 + " of component type "
                                 + self.component_name
                                 + ".")
        object.__setattr__(self, key, value)

    def _freeze(self):
        self.__isfrozen = True

    def _unfreeze(self):
        self.__isfrozen = False

    def set_AT(self, at_list, RELATIVE=None):
        """
        Method for setting position of component

        Sets the position of the component using provided length 3 list as a
        vector. If only a float is given, it is considered along the beam
        direction, [0, 0, z]. The RELATIVE keyword is used to specify in
        relation to what the position is given, and can take either a string
        with a component name or a Component object.

        Parameters
        ----------
        at_list : List of 3 floats or float
            Position of component relative to reference component

        keyword arguments:
            RELATIVE : str
                Sets reference component for position
        """
        if isinstance(at_list, (int, float, str)):
            at_list = [0, 0, at_list]

        if not isinstance(at_list, list):
            raise RuntimeError("set_AT should be given either a list or "
                               + "float, but received "
                               + str(type(at_list)))

        if len(at_list) != 3:
            raise RuntimeError("Position data given to set_AT should "
                               + "either be of length 3 or just a float.")

        # If parameter objects given, take their name instead
        for index, element in enumerate(at_list):
            if isinstance(element, (Parameter, DeclareVariable)):
                at_list[index] = element.name

        self.AT_data = at_list
        if RELATIVE is not None:
            self.set_AT_RELATIVE(RELATIVE)

    def set_AT_RELATIVE(self, relative):
        """
        Sets AT RELATIVE with string or Component instance

        Set relative which becomes the reference for this components position,
        it is possible to use a string that match another component name or
        use a Component object where the name is used as the reference.

        Parameters
        ----------

        relative : str or Component object
            Used as reference for component position
        """

        # Extract name if Component instance is given
        if isinstance(relative, Component):
            relative = relative.name
        elif not isinstance(relative, str):
            raise ValueError("Relative must be either string or "
                             + "Component object.")

        # Set AT relative
        if relative == "ABSOLUTE":
            self.AT_relative = "ABSOLUTE"
            self.AT_reference = None
        else:
            self.AT_relative = "RELATIVE " + relative
            self.AT_reference = relative

    def set_ROTATED(self, rotated_list, RELATIVE=None):
        """
        Method for setting rotation of component

        Parameters
        ----------
        rotated_list : List of 3 floats
            Rotation of component relative to reference component

        keyword arguments:
            RELATIVE : str
                Sets reference component for rotation
        """
        if not isinstance(rotated_list, list):
            raise RuntimeError("set_ROTATED should be given a list "
                               + " but received "
                               + str(type(rotated_list)))

        if len(rotated_list) != 3:
            raise RuntimeError("Rotation data given to set_ROTATED should "
                               + "be of length 3.")

        # If parameter objects given, take their name instead
        for index, element in enumerate(rotated_list):
            if isinstance(element, (Parameter, DeclareVariable)):
                rotated_list[index] = element.name

        self.ROTATED_data = rotated_list
        self.ROTATED_specified = True
        if RELATIVE is not None:
            self.set_ROTATED_RELATIVE(RELATIVE)

    def set_ROTATED_RELATIVE(self, relative):
        """
        Sets ROTATED RELATIVE with string or Component instance

        Set relative which becomes the reference for this components rotation,
        it is possible to use a string that match another component name or
        use a Component object where the name is used as the reference.

        Parameters
        ----------

        relative : str or Component object
            Used as reference for component rotation
        """

        self.ROTATED_specified = True
        # Extract name if a Component instance is given
        if isinstance(relative, Component):
            relative = relative.name
        elif not isinstance(relative, str):
            raise ValueError("Relative must be either string or "
                             + "Component object.")

        # Set ROTATED relative
        if relative == "ABSOLUTE":
            self.ROTATED_relative = "ABSOLUTE"
            self.ROTATED_reference = None
        else:
            self.ROTATED_relative = "RELATIVE " + relative
            self.ROTATED_reference = relative

    def set_RELATIVE(self, relative):
        """
        Method for setting reference of component position and rotation

        Input can be either a string matching an earlier component or a
        Component object.

        Parameters
        ----------
        relative : str
            Reference component for position and rotation
        """
        # Extract name if a Component instance is given
        if isinstance(relative, Component):
            relative = relative.name
        elif not isinstance(relative, str):
            raise ValueError("Relative must be either string or "
                             + "Component object.")

        if relative == "ABSOLUTE":
            self.AT_relative = "ABSOLUTE"
            self.AT_reference = None
            self.ROTATED_relative = "ABSOLUTE"
            self.ROTATED_reference = None
        else:
            self.AT_relative = "RELATIVE " + relative
            self.AT_reference = relative
            self.ROTATED_relative = "RELATIVE " + relative
            self.ROTATED_reference = relative

    def set_parameters(self, args_as_dict=None, **kwargs):
        """
        Set Component parameters from dictionary input or keyword arguments

        Relies on attributes added when McCode_Instr creates a subclass from
        the Component class where each component parameter is added as an
        attribute.

        An error is raised if trying to set a parameter that does not exist

        Parameters
        ----------
        args_as_dict : dict
            Parameters names and their values as dictionary
        """
        if args_as_dict is not None:
            parameter_dict = args_as_dict
        else:
            parameter_dict = kwargs

        for key, val in parameter_dict.items():
            if not hasattr(self, key):
                raise NameError("No parameter called "
                                + key
                                + " in component named "
                                + self.name
                                + " of component type "
                                + self.component_name
                                + ".")
            else:
                setattr(self, key, val)

    def set_WHEN(self, string):
        """
        Method for setting WHEN c expression of this component

        Parameters
        ----------
        WHEN : str
            Sets WHEN c expression for named McStas component
        """
        if not isinstance(string, str):
            raise RuntimeError("set_WHEN expect a string, but was "
                               + "given " + str(type(string)))
        self.WHEN = "WHEN (" + string + ")"

    def set_GROUP(self, string):
        """
        Method for setting GROUP keyword of this component

        Parameters
        ----------
        GROUP : str
            Sets GROUP name for named McStas component
        """
        if not isinstance(string, str):
            raise RuntimeError("set_GROUP expect a string, but was "
                               + "given " + str(type(string)))
        self.GROUP = string

    def set_JUMP(self, string):
        """
        Method for setting JUMP expression of this component

        Should contain all code needed after JUMP

        Parameters
        ----------
        JUMP : str
            Sets JUMP expression for named McStas component
        """
        if not isinstance(string, str):
            raise RuntimeError("set_JUMP expect a string, but was "
                               + "given " + str(type(string)))
        self.JUMP = string

    def set_SPLIT(self, value):
        """
        Method for setting SPLIT value of this component

        Parameters
        ----------
        SPLIT : int or str
            Sets SPLIT value for named McStas component
        """
        if not isinstance(value, (int, str)):
            raise RuntimeError("set_SPLIT expect a integer or string, but "
                               + "was given " + str(type(value)))

        if isinstance(value, int):
            if value < 0:
                raise RuntimeError("set_SPLIT got a negative value, this is "
                                   + "meaningless, has to be a "
                                   + "positive value.")

        self.SPLIT = value

    def append_EXTEND(self, string):
        """
        Method for adding line of c to EXTEND section of this component

        Parameters
        ----------
        EXTEND : str
            Line of c code added to EXTEND section of named component
        """
        if not isinstance(string, str):
            raise RuntimeError("append_EXTEND expect a string, but was "
                               + "given " + str(type(string)))
        self.EXTEND = self.EXTEND + string + "\n"

    def set_comment(self, string):
        """
        Sets a comment displayed before the component in written files

        Parameters
        ----------
        string : str
            Comment string
        """
        if not isinstance(string, str):
            raise RuntimeError("set_comment expect a string, but was "
                               + "given " + str(type(string)))
        self.comment = string

    def set_c_code_before(self, string):
        """
        Method for setting c code before this component

        Parameters
        ----------
        code : str
            Code to be pasted before component
        """
        if not isinstance(string, str):
            raise RuntimeError("set_c_code_before expect a string, but was "
                               + "given " + str(type(string)))
        self.c_code_before = string

    def set_c_code_after(self, string):
        """
        Method for setting c code after this component

        Parameters
        ----------
        code : str
            Code to be pasted after component
        """
        if not isinstance(string, str):
            raise RuntimeError("set_c_code_after expect a string, but was "
                               + "given " + str(type(string)))
        self.c_code_after = string

    def set_save_parameters(self, value):
        if value:
            self.save_parameters = True
        else:
            self.save_parameters = False

    def add_search(self, statement, SHELL=False):
        """
        Adds a search statement to the component

        The search line can be used to tell McStas to search for files in
        additional locations. Double quotes are added.

        Parameters
        ----------
            statement : str
                The search statement

            SHELL : bool (default False)
                if True, shell keyword is added
        """

        self.search_statement_list.add_statement(SearchStatement(statement, SHELL=SHELL))

    def clear_search(self):
        """
        Clears the component of all search statements
        """

        self.search_statement_list.clear()

    def show_search(self):
        """
        Shows all search statements of component
        """

        print(self.search_statement_list)

    def write_component(self, fo, instrument_search=None):
        """
        Method that writes component to file

        Relies on attributes added when McStas_Instr creates a subclass
        based on the component class.

        """
        parameters_per_line = 2
        # Could use character limit on lines instead
        parameters_written = 0  # internal parameter

        save_parameter_string = ""

        if len(self.c_code_before) > 0:
            explanation = "From component named " + self.name
            fo.write("%s // %s\n" % (str(self.c_code_before), explanation))
            fo.write("\n")

        # Write comment if present
        if len(self.comment) > 1:
            fo.write("// %s\n" % (str(self.comment)))

        # Write search statements
        self.search_statement_list.write(fo)

        # Add search statement for instrument if supplied
        if instrument_search is not None:
            instrument_search.write(fo)

        if self.SPLIT != 0:
            fo.write("SPLIT " + str(self.SPLIT) + " ")

        # Write component name and component type
        fo.write("COMPONENT %s = %s(" % (self.name, self.component_name))

        component_parameters = {}
        for key in self.parameter_names:
            val = getattr(self, key)
            if val is None:
                if self.parameter_defaults[key] is None:
                    raise NameError("Required parameter named "
                                    + key
                                    + " in component named "
                                    + self.name
                                    + " not set.")
                else:
                    continue
            elif isinstance(val, (Parameter, DeclareVariable)):
                # Extract the parameter name
                val = val.name

            component_parameters[key] = val

        number_of_parameters = len(component_parameters)

        if number_of_parameters == 0:
            fo.write(")\n")  # If there are no parameters, close immediately
        else:
            fo.write("\n")  # If there are parameters, start a new line

        for key, val in component_parameters.items():
            if isinstance(val, float):  # CHeck if value is a number
                # Small or large numbers written in scientific format
                fo.write(" %s = %G" % (str(key), val))
            else:
                fo.write(" %s = %s" % (str(key), str(val)))
            parameters_written = parameters_written + 1
            if parameters_written < number_of_parameters:
                fo.write(",")  # Comma between parameters
                if parameters_written % parameters_per_line == 0:
                    fo.write("\n")
            else:
                fo.write(")\n")  # End paranthesis after last parameter

        # Optional WHEN section
        if not self.WHEN == "":
            fo.write("%s\n" % self.WHEN)

        # Write AT and ROTATED section
        fo.write("AT (%s,%s,%s)" % (str(self.AT_data[0]),
                                    str(self.AT_data[1]),
                                    str(self.AT_data[2])))

        fo.write(" %s\n" % self.AT_relative)

        if self.save_parameters:
            save_parameter_string += " %s\n" % self.AT_relative

        if self.ROTATED_specified:
            fo.write("ROTATED (%s,%s,%s)" % (str(self.ROTATED_data[0]),
                                             str(self.ROTATED_data[1]),
                                             str(self.ROTATED_data[2])))

            fo.write(" %s\n" % self.ROTATED_relative)

        if not self.GROUP == "":
            fo.write("GROUP %s\n" % self.GROUP)

        # Optional EXTEND section
        if not self.EXTEND == "":
            fo.write("EXTEND %{\n")
            fo.write("%s" % self.EXTEND)
            fo.write("%}\n")

        if not self.JUMP == "":
            fo.write("JUMP %s\n" % self.JUMP)

        if len(self.c_code_after) > 0:
            fo.write("\n")
            explanation = "From component named " + self.name
            fo.write("%s // %s\n" % (str(self.c_code_after), explanation))

        # Leave a new line between components for readability
        fo.write("\n")

    def make_write_string(self):
        string = ""

        string += f'fprintf(file, "COMPONENT {self.name} = {self.component_name}\\n");\n'

        component_parameters = {}
        for key in self.parameter_names:
            val = getattr(self, key)
            if val is None:
                if self.parameter_defaults[key] is None:
                    raise NameError("Required parameter named "
                                    + key
                                    + " in component named "
                                    + self.name
                                    + " not set.")
                else:
                    continue
            elif isinstance(val, (Parameter, DeclareVariable)):
                # Extract the parameter name
                val = val.name

            component_parameters[key] = val

        for key, val in component_parameters.items():
            par_type = self.parameter_types[key] # from component reader

            cast = ""
            if par_type == "" or par_type == "double":
                type_string = "%lf"
                cast = "(double)"
            elif par_type == "int":
                type_string = "%d"
                cast = "(int)"
            elif par_type == "string":
                type_string = "%s"
            else:
                raise ValueError("Unknown parameter type: " + par_type)

            string += f'fprintf(file, "{key}={type_string}\\n", {cast} {val});\n'

        string += f'fprintf(file, "AT (%lf,%lf,%lf) {self.AT_relative}\\n",'
        string += "(double) " + str(self.AT_data[0]) + ","
        string += "(double) " + str(self.AT_data[1]) + ","
        string += "(double) " + str(self.AT_data[2]) + ");\n"

        if self.ROTATED_specified:
            string += f'fprintf(file, "ROTATED (%lf,%lf,%lf) {self.ROTATED_relative}\\n",'
            string += "(double) " + str(self.ROTATED_data[0]) + ","
            string += "(double) " + str(self.ROTATED_data[1]) + ","
            string += "(double) " + str(self.ROTATED_data[2]) + ");\n"

        return string

    def __str__(self):
        """
        Returns string of information about the component

        Includes information on required parameters if they are not yet
        specified. Information on the components are added when the
        class is used as a superclass for classes describing each
        McStas component.
        """
        string = ""

        if len(self.c_code_before) > 1:
            string += self.c_code_before + "\n"
        string += self.search_statement_list.make_string()
        if len(self.comment) > 1:
            string += "// " + self.comment + "\n"
        if self.SPLIT != 0:
            string += "SPLIT " + str(self.SPLIT) + " "
        string += "COMPONENT " + str(self.name)
        string += " = " + str(self.component_name) + "("

        last_par_name = ""
        for key in self.parameter_names:
            if getattr(self, key) is not None:
                last_par_name = key

        for key in self.parameter_names:
            val = getattr(self, key)
            parameter_name = bcolors.BOLD + key + bcolors.ENDC
            if val is not None:
                unit = ""
                if key in self.parameter_units:
                    unit += " // [" + self.parameter_units[key] + "]"
                if isinstance(val, Parameter):
                    #val_string = val.print_line() # too long
                    val_string = val.name
                elif isinstance(val, DeclareVariable):
                    val_string = val.name
                else:
                    val_string = str(val)

                value = (bcolors.BOLD
                         + bcolors.OKGREEN
                         + val_string
                         + bcolors.ENDC
                         + bcolors.ENDC)

                string += "\n"
                string += "  " + parameter_name
                string += " = " + value
                if key != last_par_name:
                    string += ","
                string += unit

            else:
                if self.parameter_defaults[key] is None:
                    string += "\n"
                    string += "  " + parameter_name
                    string += bcolors.FAIL
                    string += " : Required parameter not yet specified"
                    string += bcolors.ENDC

        string += "\n)"
        if not self.WHEN == "":
            string += " " + self.WHEN
        string += "\n"
        string += f"AT {tuple(self.AT_data)} "
        #"AT (" + str(self.AT_data[0]) + ", " + str(self.AT_data[1]) + ", " + str(self.AT_data[2]) + ") "
        string += str(self.AT_relative) + "\n"
        if self.ROTATED_specified:
            string += f"ROTATED {tuple(self.ROTATED_data)}"
            string += " " + self.ROTATED_relative + "\n"
        if not self.GROUP == "":
            string += "GROUP " + self.GROUP + "\n"
        if not self.EXTEND == "":
            string += "EXTEND %{" + "\n"
            string += self.EXTEND + "%}" + "\n"
        if not self.JUMP == "":
            string += "JUMP " + self.JUMP + "\n"
        if len(self.c_code_after) > 1:
            string += self.c_code_after + "\n"

        return string.strip()

    def print_long(self):
        """
        Prints information about the component

        Includes information on required parameters if they are not yet
        specified. Information on the components are added when the
        class is used as a superclass for classes describing each
        McStas component.
        """
        print(self.__str__())

    def __repr__(self):
        return self.__str__()

    def print_short(self, longest_name=None):
        """Prints short description of component to list print"""
        if self.ROTATED_specified:
            print_rotate_rel = self.ROTATED_relative
        else:
            print_rotate_rel = self.AT_relative

        if longest_name is not None:
            number_of_spaces = 3+longest_name-len(self.name)
            print(str(self.name) + " "*number_of_spaces, end='')
            print(str(self.component_name),
                  "\tAT", self.AT_data, self.AT_relative,
                  "ROTATED", self.ROTATED_data, print_rotate_rel)
        else:
            print(str(self.name), "=", str(self.component_name),
                  "\tAT", self.AT_data, self.AT_relative,
                  "ROTATED", self.ROTATED_data, print_rotate_rel)

    def show_parameters(self, line_length=None):
        """
        Shows available parameters and their defaults for the component

        Any value specified is not reflected in this view. The
        additional attributes defined when McStas_Instr creates
        subclasses for the individual components are required to run
        this method.
        """

        # line_limit created in _create_component_instance on instr
        if line_length is None:
            line_length = self.line_limit

        # Minimum reasonable line length
        if line_length < 74:
            line_length = 74

        print(" ___ Help "
              + self.component_name + " "
              + (line_length - 11 - len(self.component_name))*"_")
        print("|"
              + bcolors.BOLD + "optional parameter" + bcolors.ENDC + "|"
              + bcolors.BOLD
              + bcolors.UNDERLINE + "required parameter" + bcolors.ENDC
              + bcolors.ENDC + "|"
              + bcolors.BOLD
              + bcolors.OKBLUE + "default value" + bcolors.ENDC
              + bcolors.ENDC + "|"
              + bcolors.BOLD
              + bcolors.OKGREEN + "user specified value" + bcolors.ENDC
              + bcolors.ENDC + "|")

        for parameter in self.parameter_names:
            characters_before_comment = 4
            unit = ""
            if parameter in self.parameter_units:
                unit = " [" + self.parameter_units[parameter] + "]"
                characters_before_comment += len(unit)
            comment = ""
            if parameter in self.parameter_comments:
                if not self.parameter_comments[parameter] == "":
                    comment = " // " + self.parameter_comments[parameter]

            parameter_name = bcolors.BOLD + parameter + bcolors.ENDC
            characters_before_comment += len(parameter)

            value = ""
            characters_from_value = 0
            if self.parameter_defaults[parameter] is None:
                parameter_name = (bcolors.UNDERLINE
                                  + parameter_name
                                  + bcolors.ENDC)
            else:
                this_default = str(self.parameter_defaults[parameter])
                value = (" = "
                         + bcolors.BOLD
                         + bcolors.OKBLUE
                         + this_default
                         + bcolors.ENDC
                         + bcolors.ENDC)
                characters_from_value = 3 + len(this_default)

            if getattr(self, parameter) is not None:
                parameter_input = getattr(self, parameter)
                # Use name when an par/var object is found
                if isinstance(parameter_input, (Parameter, DeclareVariable)):
                    parameter_input = parameter_input.name

                this_set_value = str(parameter_input)
                value = (" = "
                         + bcolors.BOLD
                         + bcolors.OKGREEN
                         + this_set_value
                         + bcolors.ENDC
                         + bcolors.ENDC)
                characters_from_value = 3 + len(this_set_value)
            characters_before_comment += characters_from_value

            print(parameter_name + value + unit, end="")

            if characters_before_comment + len(comment) < line_length:
                print(comment)
            else:
                length_for_comment = line_length - characters_before_comment
                # Split comment into several lines
                original_comment = comment
                words = comment.split(" ")
                words_left = len(words)
                last_index = 0
                current_index = 0
                comment = ""
                iterations = 0
                max_iterations = 50
                while(words_left > 0):
                    iterations += 1
                    if iterations > max_iterations:
                        #  Something went long, print on one line
                        break

                    line_left = length_for_comment

                    while(line_left > 0):
                        if current_index >= len(words):
                            current_index = len(words) + 1
                            break
                        line_left -= len(str(words[current_index])) + 1
                        current_index += 1

                    current_index -= 1
                    for word in words[last_index:current_index]:
                        comment += word + " "
                    words_left = len(words) - current_index
                    if words_left > 0:
                        comment += "\n" + " "*characters_before_comment
                        last_index = current_index

                if not iterations == max_iterations + 1:
                    print(comment)
                else:
                    print(str(original_comment))

        print(line_length*"-")

    def show_parameters_simple(self):
        """
        Shows available parameters and their defaults for the component

        Any value specified is not reflected in this view. The
        additional attributes defined when McStas_Instr creates
        subclasses for the individual components are required to run
        this method.
        """
        print("---- Help " + self.component_name + " -----")
        for parameter in self.parameter_names:
            value = ""
            if self.parameter_defaults[parameter] is not None:
                value = " = " + str(self.parameter_defaults[parameter])
            if getattr(self, parameter) is not None:
                value = " = " + str(getattr(self, parameter))

            unit = ""
            if parameter in self.parameter_units:
                unit = " [" + self.parameter_units[parameter] + "]"

            comment = ""
            if parameter in self.parameter_comments:
                if self.parameter_comments[parameter] != "":
                    comment = " // " + self.parameter_comments[parameter]

            print(parameter + value + unit + comment)

        print("----------" + "-"*len(self.component_name) + "------")

    def check_parameters(self, instrument_variables):
        """
        Checks the current component parameters are in given name list

        The name list will typically be a list of instrument parameters and
        declared variables.

        Parameters
        ----------

        instrument_variables : list of str
            List of instrument variable names
        """
        McStas_constants = ["DEG2RAD", "RAD2DEG", "MIN2RAD", "RAD2MIN", "V2K",
                            "K2V", "VS2E", "SE2V", "FWHM2RMS", "RMS2FWHM",
                            "MNEUTRON" "HBAR", "PI", "FLT_MAX"]

        # parameter_names attribute added in _create_component_instance
        for par_name in self.parameter_names:
            par = getattr(self, par_name)
            if isinstance(par, str):
                if not par.isalpha():
                    # Allow numbers and strings with math symbols
                    # They may have errors, but can't be easily checked
                    continue

                if par in McStas_constants:
                    # Allow McStas defined constants
                    continue

                if par in instrument_variables:
                    # Allow variables on given whitelist
                    continue

                # If none of the above approves the variable, raise error
                raise McStasError("Variable not recognized.\n"
                                  f"Unrecognized variable '{par}' "
                                  "assigned to component parameter "
                                  f"'{par_name}' in component "
                                  f"'{self.name}' of type "
                                  f"'{self.component_name}'.\n"
                                  "This check can be skipped with "
                                  "settings(checks=False)")



from mcstasscript.instr_reader.util import SectionReader


class UservarsReader(SectionReader):
    """
    Reads the uservars section of a McStas instrument file and adds
    the found parameters / functions / structs to the McStasScript
    Instr instance. The information can also be written to a python
    file for reproduction of a McStas instrument.
    """

    def __init__(self, Instr, write_file, product_filename,
                 get_next_line, return_line):

        super().__init__(Instr, write_file, product_filename,
                         get_next_line, return_line)

        self.in_uservars_function = False
        self.in_struct_definition = False
        self.bracket_counter = 0

    def read_uservars_line(self, line):
        """
        Reads line of instrument uservars, returns bolean.  If it encounters
        the end of the uservars section, it returns False, otherwise True.

        The contents of the uservars section is written to the McStasScript
        Instr object.
        """

        continue_uservars = True

        # Remove comments
        if "//" in line:
            line = line.split("//", 1)[0]

        # Remove %} and signify end if this is found
        if "%}" in line:
            continue_uservars = False
            line = line.split("%}", 1)[0]

        if "/*" in line:
            line = line.split("/*", 1)[0].strip()

        if self.in_uservars_function:
            if "{" in line:
                self.bracket_counter += 1

            if "}" in line:
                self.bracket_counter -= 1

            if self.bracket_counter == 0:
                self.in_uservars_function = False

            self.Instr.append_uservars(line)
            self._write_uservars_line(line)

        # Check for functions
        if ("(" in line and ";" not in line and " " in line.strip()
                and not self.in_uservars_function):

            # If in function, it will define a block
            n_curly_brackets = line.count("{")
            n_curly_brackets -= line.count("}")

            while n_curly_brackets != 0 or ("{" not in line):
                next_line = self.get_next_line()
                line += next_line

                n_curly_brackets = line.count("{")
                n_curly_brackets -= line.count("}")

            after_curly_bracket = line.split("}")[-1]

            uservars_lines = line.split("\n")
            for uservars_line in uservars_lines:
                uservars_line = uservars_line.rstrip()
                uservars_line = uservars_line.replace('\\n', "\\\\n")
                uservars_line = uservars_line.replace('"', "\\\"")
                self.Instr.append_uservars(uservars_line)
                self._write_uservars_line(uservars_line)

            line = after_curly_bracket

        # Check for struct / function that returns struct
        if line.strip().startswith("struct "):
            # Can be a function returning struct or struct definition

            # If struct definition, no parenthesis and ; after )
            n_curly_brackets = line.count("{")
            n_curly_brackets -= line.count("}")

            # Add lines until end of block found
            while n_curly_brackets != 0 or ("{" not in line):

                next_line = self.get_next_line()
                line += next_line

                n_curly_brackets = line.count("{")
                n_curly_brackets -= line.count("}")

                if "{" in line:
                    before_curly_bracket = line.split("{", 1)[0]
                    if "(" in before_curly_bracket and ")" in before_curly_bracket:
                        # This is a function that returns a struct!
                        self.in_uservars_function = True

            after_curly_bracket = line.split("}")[-1]

            # if not in function, add until ; is found
            while ";" not in after_curly_bracket and not self.in_uservars_function:
                # It is surely a struct, find ;
                line += self.get_next_line()
                after_curly_bracket = line.split("}")[-1]

            uservars_lines = line.split("\n")
            for uservars_line in uservars_lines:
                uservars_line = uservars_line.rstrip()
                uservars_line = uservars_line.replace('\\n', "\\\\n")
                uservars_line = uservars_line.replace('"', "\\\"")
                self.Instr.append_uservars(uservars_line)
                self._write_uservars_line(uservars_line)

            if self.in_uservars_function:
                line = line.split("}")[-1].strip()
            else:
                line = line.split(";")[-1].strip()

            # if in function, stop now
            self.in_uservars_function = False

        # Grab defines
        if line.strip().startswith("#define"):
            # Include define statements as uservars append
            line = line.rstrip()
            line = line.replace('\\n', "\\\\n")
            line = line.replace('"', "\\\"")
            self.Instr.append_uservars(line)
            self._write_uservars_line(line)

        if "\n" in line:
            line = line.strip("\n")

        # Read single line parameter definitions
        if ";" in line and not self.in_uservars_function:
            # This line contains c statements
            statements = line.split(";")

            for statement in statements:
                statement = statement.strip()
                if (statement != "\n" and statement != " "
                        and len(statement) > 1):
                    self._read_uservars_statement(statement)

        return continue_uservars

    def _read_uservars_statement(self, statement):
        """
        Reads single uservars statements, which can have multiple
        variables.
        """

        statement = statement.strip()

        # Find type (same for all parameters in one statement)
        this_type = statement.split(" ", 1)[0]
        statement = statement.split(" ", 1)[1].strip()

        if this_type == "const":  # other c keywords to consider?
            this_type += " " + statement.split(" ", 1)[0]
            statement = statement.split(" ", 1)[1].strip()

        # Check for bracket initialization of arrays
        if "," in statement:
            variables = statement.split(",")
            fixed_variables = []
            array_mode = False
            for variable in variables:

                if "{" not in variable and array_mode is False:
                    fixed_variables.append(variable)
                elif "{" in variable:
                    temp_variable = variable + ","
                    array_mode = True
                elif "}" not in variable:
                    temp_variable += variable + ","
                else:
                    temp_variable += variable
                    fixed_variables.append(temp_variable)
                    array_mode = False

            variables = fixed_variables

        else:
            # No commas means just one parameter
            variables = [statement]

        # Treat each variable independently
        for variable in variables:
            variable = variable.strip()

            dynamic_size = False
            kw_args = {}

            if "=" in variable:
                value = variable.split("=")[1].strip()
                # remove the value part before proceeding
                variable = variable.split("=")[0].strip()

                if "{" in value:
                    # handle array as value
                    value = value.split("{")[1]
                    if "{" in value:
                        raise ValueError("Can not load arrays with larger"
                                         + "than 1 dimension yet.")
                    value = value.split("}")[0]
                    values = value.split(",")
                    return_value = []
                    for val in values:
                        return_value.append(float(val))
                else:
                    try:
                        return_value = float(value)
                    except:
                        value = value.replace('"', "\\\"")
                        #return_value = '"' + value + '"'
                        return_value = value

                kw_args["value"] = return_value

            # Handle array
            if "[" in variable:
                array_sizes = []
                array_size_strings = variable.split("[")
                # remove the array size part before proceeding
                variable = variable.split("[", 1)[0].strip()
                for array_size_string in array_size_strings:
                    if "]" in array_size_string:
                        this_size = array_size_string.split("]")[0]
                        try:
                            # Size uservarsd normally
                            array_sizes.append(int(this_size))
                        except:
                            # No size uservarsd means the size is automatic
                            dynamic_size = True

                if len(array_sizes) > 1:
                    raise ValueError("Can not handle arrays with larger"
                                     + " than 1 dimension yet")
                if not dynamic_size:
                    kw_args["array"] = array_sizes[0]

            if dynamic_size:
                # McStasScript needs size of array, so it is found manually
                kw_args["array"] = len(kw_args["value"])

            # value, array and typeremoved, all that remians is the name
            variable_name = variable
            self.Instr.add_user_var(this_type, variable_name, **kw_args)

            # Also write it to a file?
            write_string = []
            write_string.append(self.instr_name)
            write_string.append(".add_user_var(")
            write_string.append("\"" + this_type + "\"")
            write_string.append(", ")
            write_string.append("\"" + variable_name + "\"")
            write_string.append(self._kw_to_string(kw_args))
            write_string.append(")\n")

            # Write uservars parameter to python file
            self._write_to_file(write_string)

    def _write_uservars_line(self, string):

        string = string.rstrip()

        write_string = []
        write_string.append(self.instr_name)
        write_string.append(".append_uservars(")
        write_string.append("\"" + string + "\"")
        write_string.append(")\n")

        self._write_to_file(write_string)



import copy

import matplotlib.pyplot as plt
import numpy as np

from mcstasscript.instrument_diagram.box import ComponentBox
from mcstasscript.instrument_diagram.component_description import component_description

class DiagramCanvas:
    def __init__(self, left_side_arrows, component_boxes, right_side_arrows,
                 component_categories, colors, intensity_diagnostics=None,
                 variable=None, limits=None):
        """
        Creates diagram of instrument file with given boxes and arrows

        Makes diagram with column of boxes with arrows between them. The boxes
        are described by the ComponentBox class and arrows by Arrow class. Some
        arrows are drawn on the left side, AT and ROTATED relationships, while
        more advanced connections are on the right side. A legend is provided
        where the colors corresponding to component categories are shown along
        with the arrow colors used for different features. The right side
        arrows only show up in the legend when in use to reduce clutter. When
        using this diagram in matplotlib widget mode, it is possible get more
        information on each component by hovering the mouse of the beginning
        the component box.

        If intensity_diagnostics is given with an IntensityDiagnostics object,
        a graph of the intensity and n rays throughout the instrument is
        generated instead of the right side arrows.
        """
        self.left_side_arrows = left_side_arrows
        self.component_boxes = component_boxes
        self.right_side_arrows = right_side_arrows
        self.all_arrows = left_side_arrows + right_side_arrows
        self.component_categories = component_categories
        self.colors = colors

        self.variable = variable
        self.limits = limits
        if intensity_diagnostics is None:
            self.intensity_analysis_mode = False
        else:
            self.intensity_analysis_mode = True
            self.intensity_diagnostics = intensity_diagnostics

        # Identify cases where multiple written input goes to or from same box
        arrow_connections = {x: [] for x in self.component_boxes}
        for arrow in self.right_side_arrows:
            arrow_connections[arrow.origin].append(arrow)
            arrow_connections[arrow.target].append(arrow)

        for box in arrow_connections:
            arrows = arrow_connections[box]
            if len(arrows) > 1:  # Only look at cases with more than one arrow to / from box
                # Find unique kinds
                kinds_entry = set()
                kinds_exit = set()
                for arrow in arrows:
                    if arrow.target is box:
                        kinds_entry.add(arrow.kind)

                    if arrow.origin is box:
                        kinds_exit.add(arrow.kind)

                n_collected_arrows = len(kinds_entry) + len(kinds_exit)
                if n_collected_arrows == 1:
                    # If there is only one kind and direction, we are done
                    continue

                # Distribute placements on the box equally for the different kinds / with entry/exit kept apart
                displacements = list(np.linspace(-0.16, 0.16, n_collected_arrows))

                entry_displacements = {}
                exit_displacements = {}
                for arrow in arrows:

                    if arrow.target is box:
                        arrow.set_target_congested(True)  # Mark congested, avoids writing description on the line

                        if arrow.kind in entry_displacements:
                            arrow.set_box_offset_target(entry_displacements[arrow.kind])
                        else:
                            displacement = displacements.pop()
                            entry_displacements[arrow.kind] = displacement
                            arrow.set_box_offset_target(displacement)

                    if arrow.origin is box:
                        arrow.set_origin_congested(True) # Mark congested, avoids writing description on the line

                        if arrow.kind in exit_displacements:
                            arrow.set_box_offset_origin(exit_displacements[arrow.kind])
                        else:
                            displacement = displacements.pop()
                            exit_displacements[arrow.kind] = displacement
                            arrow.set_box_offset_origin(displacement)

        # Style constants
        self.FIG_HEIGHT_PER_BOX = 0.4
        self.FIG_WIDTH_PER_LANE = 0.33
        self.FIG_WIDTH_PER_WEIGHTED_CHARACTER = 0.122
        self.FIG_EXTRA_WIDTH_FOR_TEXT = 0.15
        self.FIG_LEFT_MARGIN = self.FIG_WIDTH_PER_LANE * 0.5
        self.FIG_RIGHT_MARGIN = self.FIG_WIDTH_PER_LANE * 0.5
        self.FIG_LEGEND_HEADLINE = 0.2
        if self.intensity_analysis_mode:
            self.FIG_RIGHT_SIDE_MINIMUM_SPACE = 5
        else:
            self.FIG_RIGHT_SIDE_MINIMUM_SPACE = 3

        self.graph_height = self.FIG_HEIGHT_PER_BOX * len(component_boxes)

        # Distribute boxes over graph height
        margin = 0.5/len(component_boxes)
        box_height_centers = np.linspace(1 - margin, margin, len(component_boxes))
        box_height = (1 - 2 * margin) / len(component_boxes)
        for box, y_pos in zip(component_boxes, box_height_centers):
            box.set_y(y_pos)
            box.set_box_height(box_height)

        lane_numbers = [x.lane for x in left_side_arrows] + [0]
        self.highest_lane_left = max(lane_numbers)

        lane_numbers = [x.lane for x in right_side_arrows] + [0]
        self.highest_lane_right = max(lane_numbers)

        weighted_box_names = [x.weighted_name_length for x in component_boxes]
        self.characters_in_longest_name = max(weighted_box_names)

        self.left_space = self.highest_lane_left * self.FIG_WIDTH_PER_LANE
        self.box_space = self.characters_in_longest_name * self.FIG_WIDTH_PER_WEIGHTED_CHARACTER + self.FIG_EXTRA_WIDTH_FOR_TEXT
        self.right_space = self.highest_lane_right * self.FIG_WIDTH_PER_LANE

        self.right_space = max(self.right_space, self.FIG_RIGHT_SIDE_MINIMUM_SPACE)

        self.graph_width = self.FIG_LEFT_MARGIN + self.left_space + self.box_space + self.right_space + self.FIG_RIGHT_MARGIN

        self.text_start_graph = (self.FIG_LEFT_MARGIN + self.left_space) / self.graph_width
        self.text_end_graph = (self.FIG_LEFT_MARGIN + self.left_space + self.box_space) / self.graph_width

        lane_width = self.left_space/self.graph_width/self.highest_lane_left

        # Arrow width and height in graph units, compensate for graph size to roughly equalize
        arrow_width = 0.02 / (self.graph_height + 0.5) # added constant to avoid asymptote near 0
        arrow_length = 0.2 / (self.graph_width + 0.05)

        # Rescale all
        for arrow in self.all_arrows:
            arrow.set_lane_width(lane_width)
            arrow.set_lane_offset(0.18)
            arrow.set_arrow_width(arrow_width)
            arrow.set_arrow_length(arrow_length)

        for box in self.component_boxes:
            box.set_box_indent(self.FIG_EXTRA_WIDTH_FOR_TEXT/self.graph_width)
            box.set_x(self.text_start_graph)  # Places boxes so they get name_width space

        # Prepare legend
        all_categories = list(set(component_categories.values()))
        if "work directory" not in all_categories:
            all_categories.append("work directory")
        all_categories.sort()

        # Colors from http://vrl.cs.brown.edu/color
        category_colors = ["#7487fb", "#81cc4c", "#e586fe", "#fe707d",
                           "#00d618", "#ffa300", "#cc9966", "#53c6ef",
                           "#baa3c6", "#aebf8a"]

        needed_colors = category_colors[:len(all_categories)]
        self.category_color_dict = dict(zip(all_categories, needed_colors))

        active_categories = []
        for box in self.component_boxes:
            if box.component_object is None:
                continue

            category = component_categories[box.component_object.component_name]
            if category not in active_categories:
                active_categories.append(category)

        self.legend_height = self.FIG_LEGEND_HEADLINE + self.FIG_HEIGHT_PER_BOX * (len(all_categories) // 2)

    def make_legend(self):

        box_top = 1 - self.FIG_LEGEND_HEADLINE / self.legend_height

        fig, ax = plt.subplots(figsize=(6.2, self.legend_height))
        ax.set(xlim=(0, 1), ylim=(0, 1))
        ax.get_xaxis().set_ticks([])
        ax.get_yaxis().set_ticks([])

        bbox = dict(boxstyle="round", facecolor="white", edgecolor="white")

        ax.text(0.385, 1.03, "Legend", va="center", ha="center", fontweight="semibold", bbox=bbox, fontsize="large")

        legend_boxes = [ComponentBox("Arm")]
        for category, color in self.category_color_dict.items():
            box = ComponentBox(category)
            box.set_background_color(color)
            legend_boxes.append(box)

        # Color components boxes after same color scheme
        for box in self.component_boxes:
            if box.component_object is None:
                continue

            component_type = box.component_object.component_name
            if component_type == "Arm":
                box.set_background_color("white")
            else:
                component_category = self.component_categories[component_type]
                box.set_background_color(self.category_color_dict[component_category])

        batch_cut = len(legend_boxes) // 2
        batches = [legend_boxes[:batch_cut], legend_boxes[batch_cut:]]

        batch_end = 0.0
        for batch in batches:
            margin = 0.5 / len(batch)
            box_height_centers = np.linspace(box_top, margin, len(batch))
            box_height = (1 - 2 * margin) / len(batch)
            for box, y_pos in zip(batch, box_height_centers):
                box.set_x(batch_end)
                box.set_y(y_pos)
                box.set_box_height(box_height)
                box.set_box_indent(self.FIG_EXTRA_WIDTH_FOR_TEXT / self.graph_width)

                box.plot_box(ax)

            fig.canvas.draw()
            for box in batch:
                box.calculate_bbox_dimensions(ax, self.graph_width)

            box_ends = [x.graph_box_end for x in batch]
            batch_end = max(box_ends)

        # Always show AT and RELATIVE, but the others only when present
        show_GROUP = False
        show_Union = False
        show_JUMP_target_index = False
        any_JUMP = False
        any_target_index = False
        for arrow in self.all_arrows:
            if arrow.kind == "GROUP":
                show_GROUP = True
            if arrow.kind == "JUMP":
                show_JUMP_target_index = True
                any_JUMP = True
            if arrow.kind == "target_index":
                show_JUMP_target_index = True
                any_target_index = True
            if arrow.kind == "Union":
                show_Union = True

        arrow_width = 0.012
        AT_HEIGHT = 0.83
        START_WIDTH = 0.47
        TEXT_WIDTH_INDENT = 0.01
        DISPLACEMENT = 0.2
        TEXT_DISPLACEMENT = 0.08
        LINE_LENGTH = 0.33

        current_displacement = 0
        ax.arrow(x=START_WIDTH, y=AT_HEIGHT + current_displacement, dx=LINE_LENGTH, dy=0, color=self.colors["AT"],
                 length_includes_head=True, width=arrow_width,
                 head_width=5.0 * arrow_width, head_length=2.5 * arrow_width)
        ax.text(START_WIDTH+TEXT_WIDTH_INDENT, AT_HEIGHT + TEXT_DISPLACEMENT + current_displacement,
                "RELATIVE AT", va="center", weight="semibold")

        current_displacement -= DISPLACEMENT
        ax.arrow(x=START_WIDTH, y=AT_HEIGHT+current_displacement, dx=LINE_LENGTH, dy=0, color=self.colors["ROTATED"],
                 length_includes_head=True, width=arrow_width,
                 head_width=5.0 * arrow_width, head_length=2.5 * arrow_width)
        ax.text(START_WIDTH + TEXT_WIDTH_INDENT, AT_HEIGHT + current_displacement + TEXT_DISPLACEMENT,
                "RELATIVE ROTATED", va="center", weight="semibold")

        if show_Union:
            current_displacement -= DISPLACEMENT
            ax.arrow(x=START_WIDTH, y=AT_HEIGHT + current_displacement, dx=LINE_LENGTH, dy=0,
                     color=self.colors["Union"], length_includes_head=True, width=arrow_width,
                     head_width=5.0 * arrow_width, head_length=2.5 * arrow_width)
            ax.text(START_WIDTH + TEXT_WIDTH_INDENT, AT_HEIGHT + current_displacement + TEXT_DISPLACEMENT,
                    "Union", va="center", weight="semibold")

        if show_JUMP_target_index:
            if any_JUMP and any_target_index:
                legend_text = "JUMP / target_index"
            elif any_JUMP:
                legend_text = "JUMP"
            elif any_target_index:
                legend_text = "target_index"

            current_displacement -= DISPLACEMENT
            ax.arrow(x=START_WIDTH, y=AT_HEIGHT+current_displacement, dx=LINE_LENGTH, dy=0,
                     color=self.colors["JUMP"], length_includes_head=True, width=arrow_width,
                     head_width=5.0 * arrow_width, head_length=2.5 * arrow_width)
            ax.text(START_WIDTH + TEXT_WIDTH_INDENT, AT_HEIGHT+current_displacement+TEXT_DISPLACEMENT,
                    legend_text, va="center", weight="semibold")

        if show_GROUP:
            current_displacement -= DISPLACEMENT
            ax.plot([START_WIDTH, START_WIDTH+LINE_LENGTH], 2*[AT_HEIGHT + current_displacement],
                    color=self.colors["GROUP"])
            ax.text(START_WIDTH + TEXT_WIDTH_INDENT, AT_HEIGHT + current_displacement + TEXT_DISPLACEMENT,
                    "GROUP", va="center", weight="semibold")

        box_y_coordinates = list(np.linspace(margin, box_top, 5))
        box_x_coordinate = 0.83

        # Check if any boxes are decorated for EXTEND or WHEN
        for box in self.component_boxes:
            if box.component_object is not None:
                if box.component_object.EXTEND != "":
                    EXTEND_box = copy.deepcopy(box)
                    EXTEND_box.name = "EXTEND"
                    EXTEND_box.background_color = "white"
                    EXTEND_box.set_x(box_x_coordinate)
                    EXTEND_box.set_y(box_y_coordinates.pop())
                    EXTEND_box.plot_box(ax)
                    break

        for box in self.component_boxes:
            if box.component_object is not None:
                if box.component_object.WHEN != "":
                    WHEN_box = copy.deepcopy(box)
                    WHEN_box.name = "WHEN"
                    WHEN_box.background_color = "white"
                    WHEN_box.set_x(box_x_coordinate)
                    WHEN_box.set_y(box_y_coordinates.pop())
                    WHEN_box.plot_box(ax)
                    break

    def plot(self):

        self.make_legend()

        fig, ax = plt.subplots(figsize=(self.graph_width, self.graph_height))
        ax.set(xlim=(0, 1), ylim=(0, 1))
        ax.axis("off")

        # Start by placing scatter points corresponding to the start of each box
        box_x = []
        box_y = []
        box_info = []
        for box in self.component_boxes:
            box_x.append(box.position_x + box.box_indent)
            box_y.append(box.position_y)
            if box.component_object is None:
                info = box.name
            else:
                info = component_description(box.component_object)
            box_info.append(info)

        # These scatter points will be the basis for mouse hovering showing annotations
        sc = ax.scatter(box_x, box_y, color="white")

        # Plot all the boxes
        for box in self.component_boxes:
            box.plot_box(ax)

        fig.canvas.draw()
        for box in self.component_boxes:
            # Calculate the box end position, requires they are already drawn
            box.calculate_bbox_dimensions(ax, self.graph_width)

        for arrow in self.left_side_arrows:
            # Plot arrows on left side
            arrow.plot_left_side(ax)

        if not self.intensity_analysis_mode:
            for arrow in self.right_side_arrows:
                # Plot arrows on right side
                arrow.plot_right_side(ax, self.text_end_graph)
        else:
            # Make insert with intensity and ray count graph

            # Find x coordinate of insert
            box_ends = [b.get_text_end() for b in self.component_boxes]
            latest_box_end = max(box_ends)
            remaining_space = 1.0 - latest_box_end
            axes_start_x = latest_box_end + 0.05 * remaining_space  # 0.1 works well if no names shown

            # Get y position for all boxes, but skip ABSOLUTE box
            y_positions = [box.position_y for box in self.component_boxes[1:]]
            y_spacing = y_positions[0] - y_positions[1]

            if self.variable is None:
                upper_y_lim = y_positions[0] + 0.5 * y_spacing
                lower_y_lim = y_positions[-1] - 0.5 * y_spacing
            else:
                upper_y_lim = y_positions[0]
                lower_y_lim = y_positions[-1]

            # Insert is done in figure coordinate system, need mother ax dimensions
            ax_pos = ax.get_position()

            # Find figure coordinates of the corners of the desired inset
            inset_y_bottom = ax_pos.y0 + lower_y_lim*(ax_pos.y1 - ax_pos.y0)
            inset_y_top = ax_pos.y0 + upper_y_lim*(ax_pos.y1 - ax_pos.y0)

            inset_x_start = ax_pos.x0 + axes_start_x*(ax_pos.x1 - ax_pos.x0)
            inset_x_end = ax_pos.x1

            # Create inset
            inset_ax = fig.add_axes((inset_x_start, inset_y_bottom,
                                     inset_x_end-inset_x_start,
                                     inset_y_top-inset_y_bottom))

            # Ensure the new axis is plotted under the old one for annotations to show up
            ax.set_zorder(4)

            # Plot graph, convey tick positions and ylimits to match main diagram
            self.intensity_diagnostics.run_general(variable=self.variable, limits=self.limits)
            self.intensity_diagnostics.plot(ax=inset_ax, fig=fig,
                                            y_tick_positions=y_positions,
                                            ylimits=[lower_y_lim, upper_y_lim],
                                            show_comp_names=False)


        # Create anotation box that will be shown when hovering the mouse over a box
        annot = ax.annotate("", xy=(0, 0), xytext=(20, 10), textcoords="offset points",
                            va="center", annotation_clip=False,
                            bbox=dict(boxstyle="round", fc="w"))
        annot.set_visible(False)

        # Helper function to update the anotation box with correct text
        def update_annot(ind):
            pos = sc.get_offsets()[ind["ind"][0]]
            annot.xy = pos

            n_lines_in_info = len(box_info[ind["ind"][0]].split("\n"))
            if n_lines_in_info > 4:
                annot.set_position((20, n_lines_in_info*(5 - pos[1]*10)))
            else:
                annot.set_position((20, 0))

            text = "{}".format(" ".join([box_info[n] for n in ind["ind"]]))
            annot.set_text(text)
            annot.get_bbox_patch().set_facecolor([0.95, 0.95, 0.95])

        # Helper function to detect hovering and update annotation accordingly
        def hover(event):
            vis = annot.get_visible()
            if event.inaxes == ax:
                cont, ind = sc.contains(event)
                if cont:
                    update_annot(ind)
                    annot.set_visible(True)
                    fig.canvas.draw_idle()
                else:
                    if vis:
                        annot.set_visible(False)
                        fig.canvas.draw_idle()

        # Connect the hover function to the canvas event signal
        fig.canvas.mpl_connect("motion_notify_event", hover)

        # Show the figure
        plt.show()


from mcstasscript.instr_reader.util import SectionReader


class InitializeReader(SectionReader):
    """
    Reads the initialize section of a McStas instrument file.
    The initialize lines are added to the McStasScript instrument, and
    are sent to the function writing the lines to the python file.
    """

    def __init__(self, Instr, write_file, product_filename,
                 get_next_line, return_line):
        super().__init__(Instr, write_file, product_filename,
                         get_next_line, return_line)

    def read_initialize_line(self, line):
        """
        Reads lines from INITIALIZE file and returns True as long as
        the stop characters has not been encountered. Comments are
        ignored with typical c syntax.
        """

        continue_initialize = True

        # Remove comments
        if "//" in line:
            line = line.split("//", 1)[0].strip()

        if line.startswith("INITIALIZE"):
            line = line.split("INITIALIZE", 1)[1].strip()

        if line.startswith("INITIALISE"):
            line = line.split("INITIALISE", 1)[1].strip()

        # Remove block opening
        if "%{" in line:
            line = line.split("%{", 1)[1].strip()

        if "%}" in line:
            line = line.split("%}", 1)[0].strip()
            continue_initialize = False

        # If the line is just a new line quit
        if line == "\n" or line == "":
            return continue_initialize

        # Remove newline at the end of the line
        if line.endswith("\n"):
            line = line[:-1]

        self.Instr.append_initialize(line)

        # Need to prepare string for being written again
        write_line = line.replace("\\n", "\\\\n")
        write_line = write_line.replace("\\t", "\\\\t")
        write_line = write_line.replace('"', '\\\"')
        # May need to expand to more cases

        # Write line to Python file
        write_string = []
        write_string.append(self.instr_name)
        write_string.append(".append_initialize(")
        write_string.append("\"" + write_line + " \"")
        write_string.append(")\n")

        self._write_to_file(write_string)

        return continue_initialize



import numpy as np
import matplotlib.pyplot as plt

from mcstasscript.helper.plot_helper import _plot_fig_ax

class EventPlotter:
    """
    Plots event data onto given views
    """
    def __init__(self, name, data, flag_info=None):
        """
        EventPlotter stores name and data, can produce plots given views

        Parameters:

        name : str
            Name of dataset

        data : McStasEventData
            Data object with event data

        flag_info : list of str
            List of flag names in order U1 U2 U3
        """
        self.name = name
        self.data = data
        self.flag_info = flag_info

    def scale_weights(self, factor):
        """
        Scales all weights in contained McStasEventData object

        Parameters:

        factor : float
            Scale factor to be applied
        """
        index = self.data.find_variable_index("p", flag_info=self.flag_info)
        self.data.Events[:, index] *= factor

    def add_view_limits(self, view):
        """
        Sets limits of View object from what is applicable to this data

        Parameters:
        view : View
            View for which limits should be set
        """
        data = self.data.get_data_column(view.axis1, flag_info=self.flag_info)
        view.set_axis1_limits(np.min(data), np.max(data))

        if view.axis2 is not None:
            data = self.data.get_data_column(view.axis1, flag_info=self.flag_info)
            view.set_axis1_limits(np.min(data), np.max(data))

    def get_view_limits_axis1(self, view):
        """
        Finds limits for axis1 of given view with the contained data

        Parameters:
        view : View
            View for which limits should be retrieved
        """
        data = self.data.get_data_column(view.axis1, flag_info=self.flag_info)
        return np.min(data), np.max(data)

    def get_view_limits_axis2(self, view):
        """
        Finds limits for axis1 of given view with the contained data

        Parameters:
        view : View
            View for which limits should be retrieved
        """
        if view.axis2 is None:
            return np.NaN, np.NaN
        data = self.data.get_data_column(view.axis2, flag_info=self.flag_info)
        return np.min(data), np.max(data)

    def plot(self, view, fig, ax):
        """
        Plots binned data generated from contained data on axis from view

        A view can also contain additional plot options which will be passed
        to the standard plotting tools.

        view : View
            View defining onto which axis and what bins EventData should be binned

        fig : Matplotlib fig
            Figure object for figure

        ax : Matplotlib ax
            Axes object for figure
        """

        if view.axis2 is None:
            data = self.data.make_1d(axis1=view.axis1, n_bins=view.bins, flag_info=self.flag_info)
            data.set_title("")
            if view.axis1_limits is not None:
                data.set_plot_options(left_lim=view.axis1_limits[0], right_lim=view.axis1_limits[1])
            data.set_plot_options(**view.plot_options)

        else:
            data = self.data.make_2d(axis1=view.axis1, axis2=view.axis2, n_bins=view.bins, flag_info=self.flag_info)
            data.set_plot_options(show_colorbar=False)
            data.set_title("")
            if view.axis1_limits is not None and view.axis2_limits is not None:
                data.set_plot_options(left_lim=view.axis1_limits[0], right_lim=view.axis1_limits[1],
                                      bottom_lim=view.axis2_limits[0], top_lim=view.axis2_limits[1])
            data.set_plot_options(**view.plot_options)

        _plot_fig_ax(data, fig, ax)

        x_lims = ax.get_xlim()
        y_lims = ax.get_ylim()

        if view.axis1_values is not None:
            for value in view.axis1_values:
                ax.plot([value, value], y_lims, "k")

            ax.set_xlim(x_lims)
            ax.set_ylim(y_lims)

        if view.axis2_values is not None:
            for value in view.axis2_values:
                ax.plot(x_lims, [value, value], "k")

            ax.set_xlim(x_lims)
            ax.set_ylim(y_lims)




import os
import unittest
import unittest.mock

from mcstasscript.helper.managed_mcrun import ManagedMcrun
from mcstasscript.helper.managed_mcrun import load_results
from mcstasscript.helper.managed_mcrun import load_metadata
from mcstasscript.helper.managed_mcrun import load_monitor
from mcstasscript.tests.helpers_for_tests import WorkInTestDir

class TestManagedMcrun(unittest.TestCase):
    """
    Testing the ManagedMcrun class that sets up McStas runs, runs the
    simulation and loads the data.

    Here the simulation is not actually performed, this will be done in
    integration tests. The surrounding plumbing and data loading is
    tested.
    """

    def test_ManagedMcrun_init_simple(self):
        """
        Check shortest possible initialization works
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_folder",
                                     executable_path=executable_path,
                                     executable="mcrun")

        self.assertEqual(mcrun_obj.name_of_instrumentfile, "test.instr")

        expected_data_folder = os.path.join(THIS_DIR, "test_folder")
        self.assertEqual(mcrun_obj.data_folder_name, expected_data_folder)

        expected_executable_path = os.path.join(THIS_DIR, "dummy_mcstas")
        self.assertEqual(mcrun_obj.executable_path, expected_executable_path)

    def test_ManagedMcrun_init_defaults(self):
        """
        Check default values are set up correctly
        """
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_folder",
                                     executable_path=executable_path,
                                     executable="mcrun")

        self.assertEqual(mcrun_obj.mpi, None)
        self.assertEqual(mcrun_obj.ncount, 1000000)
        expected_run_path = os.path.join(THIS_DIR, ".")
        self.assertEqual(mcrun_obj.run_path, expected_run_path)

    def test_ManagedMcrun_init_set_values(self):
        """
        Check values given to ManagedMcrun are internalized

        run_path set to an existing folder in the test directory
        """
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_folder",
                                     executable="mcrun",
                                     executable_path=executable_path,
                                     run_path="test_data_set",
                                     mpi=4,
                                     ncount=128)

        self.assertEqual(mcrun_obj.mpi, 4)
        self.assertEqual(mcrun_obj.ncount, 128)
        expected_run_path = os.path.join(THIS_DIR, "test_data_set")
        self.assertEqual(mcrun_obj.run_path, expected_run_path)

    def test_ManagedMcrun_init_set_parameters(self):
        """
        Check parameters can be given as dictionary
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        par_input = {"A_par": 5.1,
                     "int_par": 1,
                     "define_par": "Bike",
                     "string_par": "\"Car\""}

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_folder",
                                     executable_path=executable_path,
                                     executable="",
                                     parameters=par_input)

        self.assertEqual(mcrun_obj.parameters["A_par"], 5.1)
        self.assertEqual(mcrun_obj.parameters["int_par"], 1)
        self.assertEqual(mcrun_obj.parameters["define_par"], "Bike")
        self.assertEqual(mcrun_obj.parameters["string_par"], "\"Car\"")

    def test_ManagedMcrun_init_set_custom_flags(self):
        """
        Check custom_flags can be given by user
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        custom_flag_input = "-p"
        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_folder",
                                     executable="mcrun",
                                     executable_path=executable_path,
                                     custom_flags=custom_flag_input)

        self.assertEqual(mcrun_obj.custom_flags, custom_flag_input)

    def test_ManagedMcrun_init_no_folder_error(self):
        """
        An error should occur if no filename is given
        """
        with self.assertRaises(NameError):
            ManagedMcrun("test.instr", mcrun_path="")

    def test_ManagedMcrun_init_invalid_ncount_error(self):
        """
        An error should occur if negative ncount is given
        """
        with self.assertRaises(ValueError):
            ManagedMcrun("test.instr",
                         output_path="test_folder",
                         mcrun_path="",
                         ncount=-8)

    def test_ManagedMcrun_init_invalid_mpi_error(self):
        """
        An error should occur if negative mpi is given
        """
        with self.assertRaises(ValueError):
            ManagedMcrun("test.instr",
                         output_path="test_folder",
                         mcrun_path="",
                         mpi=-8)

    def test_ManagedMcrun_init_invalid_parameters_error(self):
        """
        An error should occur if parameters is given as non dict
        """
        with self.assertRaises(RuntimeError):
            ManagedMcrun("test.instr",
                         output_path="test_folder",
                         mcrun_path="",
                         parameters=[1, 2, 3])

    @unittest.mock.patch("subprocess.run")
    def test_ManagedMcrun_run_simulation_basic(self, mock_sub):
        """
        Check a basic system call is correct
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_folder",
                                     executable_path=executable_path,
                                     executable="mcrun",)

        mcrun_obj.run_simulation()

        expected_folder_path = os.path.join(THIS_DIR, "test_folder")

        # a double space because of a missing option
        executable = os.path.join(executable_path, "mcrun")
        executable = '"' + executable + '"'

        expected_call = (executable + " -c -n 1000000 "
                         + "-d " + expected_folder_path + "  test.instr")

        mock_sub.assert_called_once_with(expected_call,
                                         shell=True,
                                         stderr=-2, stdout=-1,
                                         universal_newlines=True,
                                         cwd=mcrun_obj.run_path)

    @unittest.mock.patch("subprocess.run")
    def test_ManagedMcrun_run_simulation_basic_path(self, mock_sub):
        """
        Check a basic system call is correct, with different path format
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas", "")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_folder",
                                     executable_path=executable_path,
                                     executable="mcrun",)

        mcrun_obj.run_simulation()

        expected_folder_path = os.path.join(THIS_DIR, "test_folder")

        executable = os.path.join(executable_path, "mcrun")
        executable = '"' + executable + '"'
        # a double space because of a missing option
        expected_call = (executable + " -c -n 1000000 "
                         + "-d " + expected_folder_path + "  test.instr")

        mock_sub.assert_called_once_with(expected_call,
                                         shell=True,
                                         stderr=-2, stdout=-1,
                                         universal_newlines=True,
                                         cwd=mcrun_obj.run_path)

    @unittest.mock.patch("subprocess.run")
    def test_ManagedMcrun_run_simulation_no_standard(self, mock_sub):
        """
        Check a non standard system call is correct

        Here multiple options are used and ncount is a float that should
        be rounded by the class.
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_folder",
                                     executable_path=executable_path,
                                     executable="mcrun",
                                     mpi=7, seed=300,
                                     ncount=48.4,
                                     custom_flags="-fo")

        mcrun_obj.run_simulation()

        expected_folder_path = os.path.join(THIS_DIR, "test_folder")

        # a double space because of a missing option
        executable = os.path.join(executable_path, "mcrun")
        executable = '"' + executable + '"'
        expected_call = (executable + " -c -n 48 --mpi=7 --seed=300 "
                         + "-d " + expected_folder_path + " -fo test.instr")

        mock_sub.assert_called_once_with(expected_call,
                                         shell=True,
                                         stderr=-2, stdout=-1,
                                         universal_newlines=True,
                                         cwd=mcrun_obj.run_path)

    @unittest.mock.patch("subprocess.run")
    def test_ManagedMcrun_run_simulation_with_gravity(self, mock_sub):
        """
        Check a non standard system call is correct when including gravity

        Here multiple options are used and ncount is a float that should
        be rounded by the class.
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_folder",
                                     executable_path=executable_path,
                                     executable="mcrun",
                                     gravity=True,
                                     mpi=7, seed=300,
                                     ncount=48.4,
                                     custom_flags="-fo")

        mcrun_obj.run_simulation()

        expected_folder_path = os.path.join(THIS_DIR, "test_folder")

        # a double space because of a missing option
        executable = os.path.join(executable_path, "mcrun")
        executable = '"' + executable + '"'
        expected_call = (executable + " -c -g -n 48 --mpi=7 --seed=300 "
                         + "-d " + expected_folder_path + " -fo test.instr")

        mock_sub.assert_called_once_with(expected_call,
                                         shell=True,
                                         stderr=-2, stdout=-1,
                                         universal_newlines=True,
                                         cwd=mcrun_obj.run_path)

    @unittest.mock.patch("subprocess.run")
    def test_ManagedMcrun_run_simulation_parameters(self, mock_sub):
        """
        Check a run with parameters is correct
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_folder",
                                     executable_path=executable_path,
                                     executable="mcrun",
                                     mpi=7,
                                     ncount=48.4,
                                     custom_flags="-fo",
                                     parameters={"A": 2,
                                                 "BC": "car",
                                                 "th": "\"toy\""})

        mcrun_obj.run_simulation()

        expected_folder_path = os.path.join(THIS_DIR, "test_folder")
        # a double space because of a missing option
        executable = os.path.join(executable_path, "mcrun")
        executable = '"' + executable + '"'

        expected_call = (executable + " -c -n 48 --mpi=7 "
                         + "-d " + expected_folder_path + " -fo test.instr "
                         + "A=2 BC=car th=\"toy\"")

        mock_sub.assert_called_once_with(expected_call,
                                         shell=True,
                                         stderr=-2, stdout=-1,
                                         universal_newlines=True,
                                         cwd=mcrun_obj.run_path)

    @unittest.mock.patch("subprocess.run")
    def test_ManagedMcrun_run_simulation_compile(self, mock_sub):
        """
        Check run with force_compile set to False works
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_folder",
                                     executable_path=executable_path,
                                     executable="mcrun",
                                     mpi=7,
                                     ncount=48.4,
                                     force_compile=False,
                                     custom_flags="-fo",
                                     parameters={"A": 2,
                                                 "BC": "car",
                                                 "th": "\"toy\""})

        mcrun_obj.run_simulation()

        expected_folder_path = os.path.join(THIS_DIR, "test_folder")

        executable = os.path.join(executable_path, "mcrun")
        executable = '"' + executable + '"'
        # a double space because of a missing option
        expected_call = (executable + " -n 48 --mpi=7 "
                         + "-d " + expected_folder_path + " -fo test.instr "
                         + "A=2 BC=car th=\"toy\"")

        mock_sub.assert_called_once_with(expected_call,
                                         shell=True,
                                         stderr=-2, stdout=-1,
                                         universal_newlines=True,
                                         cwd=mcrun_obj.run_path)

    @unittest.mock.patch("subprocess.run")
    def test_ManagedMcrun_run_simulation_NeXus(self, mock_sub):
        """
        Check run with NeXus works
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_folder",
                                     executable_path=executable_path,
                                     executable="mcrun",
                                     mpi=7,
                                     ncount=57.4,
                                     force_compile=False,
                                     NeXus=True,
                                     custom_flags="-fo",
                                     parameters={"A": 2,
                                                 "BC": "car",
                                                 "th": "\"toy\""})

        mcrun_obj.run_simulation()

        expected_folder_path = os.path.join(THIS_DIR, "test_folder")

        executable = os.path.join(executable_path, "mcrun")
        executable = '"' + executable + '"'
        # a double space because of a missing option
        expected_call = (executable + " --format=NeXus -n 57 --mpi=7 "
                         + "-d " + expected_folder_path + " -fo test.instr "
                         + "A=2 BC=car th=\"toy\"")

        mock_sub.assert_called_once_with(expected_call,
                                         shell=True,
                                         stderr=-2, stdout=-1,
                                         universal_newlines=True,
                                         cwd=mcrun_obj.run_path)

    @unittest.mock.patch("subprocess.run")
    def test_ManagedMcrun_run_simulation_openacc(self, mock_sub):
        """
        Check run with openacc works
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_folder",
                                     executable_path=executable_path,
                                     executable="mcrun",
                                     mpi=7,
                                     ncount=48.4,
                                     openacc=True,
                                     force_compile=False,
                                     custom_flags="-fo",
                                     parameters={"A": 2,
                                                 "BC": "car",
                                                 "th": "\"toy\""})

        mcrun_obj.run_simulation()

        expected_folder_path = os.path.join(THIS_DIR, "test_folder")

        executable = os.path.join(executable_path, "mcrun")
        executable = '"' + executable + '"'
        # a double space because of a missing option
        expected_call = (executable + " --openacc -n 48 --mpi=7 "
                         + "-d " + expected_folder_path + " -fo test.instr "
                         + "A=2 BC=car th=\"toy\"")

        mock_sub.assert_called_once_with(expected_call,
                                         shell=True,
                                         stderr=-2, stdout=-1,
                                         universal_newlines=True,
                                         cwd=mcrun_obj.run_path)

    def test_ManagedMcrun_load_data_PSD4PI(self):
        """
        Use test_data_set to test load_data for PSD_4PI

        test_data_set contains three data files and some junk, the mccode.sim
        file contains names of the data files so only these are loaded.
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_data_set",
                                     executable_path=executable_path,
                                     mcrun_path="path")

            results = mcrun_obj.load_results()

        # Check three data objects are loaded
        self.assertEqual(len(results), 4)

        # Check properties of PSD_4PI data
        PSD_4PI = results[0]

        self.assertEqual(PSD_4PI.name, "PSD_4PI")
        self.assertEqual(PSD_4PI.metadata.dimension, [300, 300])
        self.assertEqual(PSD_4PI.metadata.limits, [-180, 180, -90, 90])
        self.assertEqual(PSD_4PI.metadata.xlabel, "Longitude [deg]")
        self.assertEqual(PSD_4PI.metadata.ylabel, "Latitude [deg]")
        self.assertEqual(PSD_4PI.metadata.title, "4PI PSD monitor")
        expected_parameters = {"wavelength": 1.0}
        self.assertEqual(PSD_4PI.metadata.info["Parameters"], expected_parameters)
        self.assertEqual(PSD_4PI.metadata.parameters, expected_parameters)
        self.assertEqual(PSD_4PI.Ncount[4][1], 4)
        self.assertEqual(PSD_4PI.Intensity[4][1], 1.537334562E-10)
        self.assertEqual(PSD_4PI.Error[4][1], 1.139482296E-10)

    def test_ManagedMcrun_load_data_PSD(self):
        """
        Use test_data_set to test load_data for PSD
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_data_set",
                                     executable_path=executable_path,
                                     mcrun_path="path")

            results = mcrun_obj.load_results()

        # Check three data objects are loaded
        self.assertEqual(len(results), 4)

        # Check properties of PSD data
        PSD = results[1]

        self.assertEqual(PSD.name, "PSD")
        self.assertEqual(PSD.metadata.dimension, [200, 200])
        self.assertEqual(PSD.metadata.limits, [-5, 5, -5, 5])
        self.assertEqual(PSD.metadata.xlabel, "X position [cm]")
        self.assertEqual(PSD.metadata.ylabel, "Y position [cm]")
        self.assertEqual(PSD.metadata.title, "PSD monitor")
        expected_parameters = {"wavelength": 1.0}
        self.assertEqual(PSD.metadata.info["Parameters"], expected_parameters)
        self.assertEqual(PSD.metadata.parameters, expected_parameters)
        self.assertEqual(PSD.Ncount[27][21], 9)
        self.assertEqual(PSD.Intensity[27][21], 2.623929371e-13)
        self.assertEqual(PSD.Error[27][21], 2.765467693e-13)

    def test_ManagedMcrun_load_data_L_mon(self):
        """
        Use test_data_set to test load_data for L_mon
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_data_set",
                                     executable_path=executable_path,
                                     mcrun_path="path")

            results = mcrun_obj.load_results()

        # Check three data objects are loaded
        self.assertEqual(len(results), 4)

        # Check properties of L_mon
        L_mon = results[2]

        self.assertEqual(L_mon.name, "L_mon")
        self.assertEqual(L_mon.metadata.dimension, 150)
        self.assertEqual(L_mon.metadata.limits, [0.7, 1.3])
        self.assertEqual(L_mon.metadata.xlabel, "Wavelength [AA]")
        self.assertEqual(L_mon.metadata.ylabel, "Intensity")
        self.assertEqual(L_mon.metadata.title, "Wavelength monitor")
        expected_parameters = {"wavelength": 1.0}
        self.assertEqual(L_mon.metadata.info["Parameters"], expected_parameters)
        self.assertEqual(L_mon.metadata.parameters, expected_parameters)
        self.assertEqual(L_mon.xaxis[53], 0.914)
        self.assertEqual(L_mon.Ncount[53], 37111)
        self.assertEqual(L_mon.Intensity[53], 6.990299315e-06)
        self.assertEqual(L_mon.Error[53], 6.215308587e-08)

    def test_ManagedMcrun_load_data_L_mon_direct(self):
        """
        Use test_data_set to test load_data for L_mon with direct path
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_data_set",
                                     executable_path=executable_path,
                                     mcrun_path="path")

            load_path = os.path.join(THIS_DIR, "test_data_set")
            results = mcrun_obj.load_results(load_path)

        # Check properties of L_mon
        L_mon = results[2]

        self.assertEqual(L_mon.name, "L_mon")
        self.assertEqual(L_mon.metadata.dimension, 150)
        self.assertEqual(L_mon.metadata.limits, [0.7, 1.3])
        self.assertEqual(L_mon.metadata.xlabel, "Wavelength [AA]")
        self.assertEqual(L_mon.metadata.ylabel, "Intensity")
        self.assertEqual(L_mon.metadata.title, "Wavelength monitor")
        expected_parameters = {"wavelength": 1.0}
        self.assertEqual(L_mon.metadata.info["Parameters"], expected_parameters)
        self.assertEqual(L_mon.metadata.parameters, expected_parameters)
        self.assertEqual(L_mon.xaxis[53], 0.914)
        self.assertEqual(L_mon.Ncount[53], 37111)
        self.assertEqual(L_mon.Intensity[53], 6.990299315e-06)
        self.assertEqual(L_mon.Error[53], 6.215308587e-08)

        self.assertFalse(hasattr(L_mon, 'Events'))

    def test_ManagedMcrun_load_data_Event(self):
        """
        Use test_data_set to test load_data for event data
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_data_set",
                                     executable_path=executable_path,
                                     mcrun_path="path")

            load_path = os.path.join(THIS_DIR, "test_data_set")
            results = mcrun_obj.load_results(load_path)

        # Check properties of event data file
        mon = results[3]

        self.assertEqual(mon.name, "monitor")
        self.assertEqual(mon.metadata.dimension, [8, 12000])
        self.assertEqual(mon.metadata.limits, [1.0, 12000.0, 1.0, 8.0])
        self.assertEqual(mon.metadata.xlabel, "List of neutron events")
        self.assertEqual(mon.metadata.ylabel, "p x y z vx vy vz t")
        self.assertEqual(mon.metadata.title, "Intensity Position Position"
                + " Position Velocity Velocity Velocity"
                + " Time_Of_Flight Monitor (Square)")
        expected_parameters = {"wavelength": 1.0}
        self.assertEqual(mon.metadata.info["Parameters"], expected_parameters)
        self.assertEqual(mon.metadata.parameters, expected_parameters)
        self.assertEqual(mon.Events[12, 1], -0.006163896406)
        self.assertEqual(mon.Events[43, 4], 22.06193582)

        self.assertFalse(hasattr(mon, 'xaxis'))
        self.assertFalse(hasattr(mon, 'Error'))
        self.assertFalse(hasattr(mon, 'Ncount'))

    def test_ManagedMcrun_load_data_nonexisting(self):
        """
        If folder does not exists, a warning should be shown and None returned
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_data_set",
                                     executable_path=executable_path,
                                     mcrun_path="path")

        load_path = os.path.join(THIS_DIR, "non_existent_dataset")

        with self.assertWarns(Warning):
            result = mcrun_obj.load_results(load_path)

        self.assertIsNone(result)

    def test_ManagedMcrun_load_data_no_mcsim_file(self):
        """
        Check an error occurs when pointed to directory without mcsim file
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        executable_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with WorkInTestDir() as handler:
            mcrun_obj = ManagedMcrun("test.instr",
                                     output_path="test_data_set",
                                     executable_path=executable_path,
                                     mcrun_path="path")

            load_path = os.path.join(THIS_DIR, "dummy_mcstas")

        with self.assertRaises(NameError):
            mcrun_obj.load_results(load_path)


class Test_load_functions(unittest.TestCase):
    """
    Testing the load functions in managed_mcrun.
    load_results loads all data in folder
    load_metadata loads all metadata in folder
    load_monitor loads one monitor given metadata and folder
    These are used in ManagedMcrun
    """
    def test_mcrun_load_data_PSD4PI(self):
        """
        Use test_data_set to test load_data for PSD_4PI
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        results = load_results("test_data_set")

        os.chdir(current_work_dir)  # Reset work directory

        self.assertEqual(len(results), 4)

        PSD_4PI = results[0]

        self.assertEqual(PSD_4PI.name, "PSD_4PI")
        self.assertEqual(PSD_4PI.metadata.dimension, [300, 300])
        self.assertEqual(PSD_4PI.metadata.limits, [-180, 180, -90, 90])
        self.assertEqual(PSD_4PI.metadata.xlabel, "Longitude [deg]")
        self.assertEqual(PSD_4PI.metadata.ylabel, "Latitude [deg]")
        self.assertEqual(PSD_4PI.metadata.title, "4PI PSD monitor")
        expected_parameters = {"wavelength": 1.0}
        self.assertEqual(PSD_4PI.metadata.info["Parameters"], expected_parameters)
        self.assertEqual(PSD_4PI.metadata.parameters, expected_parameters)
        self.assertEqual(PSD_4PI.Ncount[4][1], 4)
        self.assertEqual(PSD_4PI.Intensity[4][1], 1.537334562E-10)
        self.assertEqual(PSD_4PI.Error[4][1], 1.139482296E-10)

    def test_mcrun_load_data_PSD(self):
        """
        Use test_data_set to test load_data for PSD
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        results = load_results("test_data_set")

        os.chdir(current_work_dir)  # Reset work directory

        self.assertEqual(len(results), 4)

        PSD = results[1]

        self.assertEqual(PSD.name, "PSD")
        self.assertEqual(PSD.metadata.dimension, [200, 200])
        self.assertEqual(PSD.metadata.limits, [-5, 5, -5, 5])
        self.assertEqual(PSD.metadata.xlabel, "X position [cm]")
        self.assertEqual(PSD.metadata.ylabel, "Y position [cm]")
        self.assertEqual(PSD.metadata.title, "PSD monitor")
        expected_parameters = {"wavelength": 1.0}
        self.assertEqual(PSD.metadata.info["Parameters"], expected_parameters)
        self.assertEqual(PSD.metadata.parameters, expected_parameters)
        self.assertEqual(PSD.Ncount[27][21], 9)
        self.assertEqual(PSD.Intensity[27][21], 2.623929371e-13)
        self.assertEqual(PSD.Error[27][21], 2.765467693e-13)

    def test_mcrun_load_metadata_PSD4PI(self):
        """
        Use test_data_set to test load_metadata for PSD_4PI
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        metadata = load_metadata("test_data_set")

        os.chdir(current_work_dir)  # Reset work directory

        self.assertEqual(len(metadata), 4)

        PSD_4PI = metadata[0]
        self.assertEqual(PSD_4PI.dimension, [300, 300])
        self.assertEqual(PSD_4PI.limits, [-180, 180, -90, 90])
        self.assertEqual(PSD_4PI.xlabel, "Longitude [deg]")
        self.assertEqual(PSD_4PI.ylabel, "Latitude [deg]")
        self.assertEqual(PSD_4PI.title, "4PI PSD monitor")
        expected_parameters = {"wavelength": 1.0}
        self.assertEqual(PSD_4PI.info["Parameters"], expected_parameters)
        self.assertEqual(PSD_4PI.parameters, expected_parameters)

    def test_mcrun_load_metadata_L_mon(self):
        """
        Use test_data_set to test load_metadata for PSD_4PI
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        metadata = load_metadata("test_data_set")

        os.chdir(current_work_dir)  # Reset work directory

        self.assertEqual(len(metadata), 4)

        L_mon = metadata[2]
        self.assertEqual(L_mon.dimension, 150)
        self.assertEqual(L_mon.limits, [0.7, 1.3])
        self.assertEqual(L_mon.xlabel, "Wavelength [AA]")
        self.assertEqual(L_mon.ylabel, "Intensity")
        self.assertEqual(L_mon.title, "Wavelength monitor")
        expected_parameters = {"wavelength": 1.0}
        self.assertEqual(L_mon.info["Parameters"], expected_parameters)
        self.assertEqual(L_mon.parameters, expected_parameters)

    def test_mcrun_load_monitor_PSD4PI(self):
        """
        Use test_data_set to test load_monitor for PSD_4PI
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        metadata = load_metadata("test_data_set")
        PSD_4PI = metadata[0]
        monitor = load_monitor(PSD_4PI, "test_data_set")

        os.chdir(current_work_dir)  # Reset work directory

        self.assertEqual(monitor.name, "PSD_4PI")
        self.assertEqual(monitor.metadata.dimension, [300, 300])
        self.assertEqual(monitor.metadata.limits, [-180, 180, -90, 90])
        self.assertEqual(monitor.metadata.xlabel, "Longitude [deg]")
        self.assertEqual(monitor.metadata.ylabel, "Latitude [deg]")
        self.assertEqual(monitor.metadata.title, "4PI PSD monitor")
        expected_parameters = {"wavelength": 1.0}
        self.assertEqual(monitor.metadata.info["Parameters"], expected_parameters)
        self.assertEqual(monitor.metadata.parameters, expected_parameters)
        self.assertEqual(monitor.Ncount[4][1], 4)
        self.assertEqual(monitor.Intensity[4][1], 1.537334562E-10)
        self.assertEqual(monitor.Error[4][1], 1.139482296E-10)

    def test_mcrun_load_monitor_L_mon(self):
        """
        Use test_data_set to test load_monitor for PSD_4PI
        """

        THIS_DIR = os.path.dirname(os.path.abspath(__file__))

        current_work_dir = os.getcwd()
        os.chdir(THIS_DIR)  # Set work directory to test folder

        metadata = load_metadata("test_data_set")
        L_mon = metadata[2]
        monitor = load_monitor(L_mon, "test_data_set")

        os.chdir(current_work_dir)  # Reset work directory

        self.assertEqual(monitor.name, "L_mon")
        self.assertEqual(monitor.metadata.dimension, 150)
        self.assertEqual(monitor.metadata.limits, [0.7, 1.3])
        self.assertEqual(monitor.metadata.xlabel, "Wavelength [AA]")
        self.assertEqual(monitor.metadata.ylabel, "Intensity")
        self.assertEqual(monitor.metadata.title, "Wavelength monitor")
        expected_parameters = {"wavelength": 1.0}
        self.assertEqual(monitor.metadata.info["Parameters"], expected_parameters)
        self.assertEqual(monitor.metadata.parameters, expected_parameters)
        self.assertEqual(monitor.xaxis[53], 0.914)
        self.assertEqual(monitor.Ncount[53], 37111)
        self.assertEqual(monitor.Intensity[53], 6.990299315e-06)
        self.assertEqual(monitor.Error[53], 6.215308587e-08)

if __name__ == '__main__':
    unittest.main()



from mcstasscript.instrument_diagram.connections import ConnectionList
from mcstasscript.instrument_diagram.arrow import Arrow


def generate_target_index_arrows(components, component_box_dict, box_names, color=None):
    """
    Generate Arrow objects related to the target_index
    """
    connections = ConnectionList()
    component_names = [x.name for x in components]

    for component in components:
        if not hasattr(component, "target_index"):
            # Component doesnt have the target_index setting
            continue
        if component.target_index is None:
            # A value has not been specified for target_index setting
            continue
        if component.target_index == 0:
            # target_index is disabled
            continue

        try:
            int(component.target_index)
        except:
            # Skip cases where target_index is not an integer
            continue

        this_component_index = component_names.index(component.name)
        target_component_index = this_component_index + int(component.target_index)
        target_component_reference = component_names[target_component_index]

        if target_component_reference not in component_box_dict:
            raise ValueError("target_index reference: "
                             + str(target_component_reference)
                             + " not found.")

        origin = component_box_dict[component.name]
        connections.add(origin, component_box_dict[target_component_reference])

    connections.distribute_lane_numbers(box_names=box_names)

    arrows = []
    for connection in connections.get_connections():
        origin = connection.origin
        target = connection.target
        lane = connection.lane_number + 3  # Make room for target_index

        arrow = Arrow(origin, target, lane=lane, kind="target_index", description="target_index")
        arrow.set_sub_lane(2)
        if color is None:
            arrow.color = "black"
        else:
            arrow.color = color

        arrows.append(arrow)

    return arrows



class View:
    def __init__(self, axis1, axis2=None, bins=100, same_scale=False,
                 axis1_values=None, axis2_values=None, **kwargs):
        """
        Plot view on one or two axis to be generated from event data

        The possible specifiers are defined in McStasEvenData and are
        x : x position [m]
        y : x position [m]
        z : x position [m]
        vx : x velocity [m/s]
        vy : x velocity [m/s]
        vz : x velocity [m/s]
        speed : [m/s]
        dx : divergence x [deg]
        dy : divergence y [deg]
        t : time [s]
        l : wavelength [AA]
        e : energy [meV]

        Parameters:

        axis1 : str
            Specifier for first axis to be shown

        axis2 : str (optional)
            Specifier for second axis to be shown

        bins : int or [int, int]
            Number of bins for generation of histogram

        axis1_values : float or list of floats
            Values that will be plotted as lines on plot

        axis2_values : float or list of floats
            Values that will be plotted as lines on plot

        same_scale : bool
            Controls whether all displays of this view is on same ranges
        """
        self.same_scale = same_scale

        self.axis1 = axis1
        self.axis1_limits = None

        if isinstance(axis1_values, (float, int)):
            axis1_values = [axis1_values]
        self.axis1_values = axis1_values

        self.axis2 = axis2
        self.axis2_limits = None
        if isinstance(axis2_values, (float, int)):
            axis2_values = [axis2_values]
        self.axis2_values = axis2_values

        self.bins = bins

        self.plot_options = kwargs

    def __repr__(self):
        string = f"View ({self.axis1}"
        if self.axis2 is not None:
            string += f", {self.axis2}"

        string += ")"
        string = string.ljust(25)
        string += f" bins: {str(self.bins)}"

        return string

    def set_axis1_limits(self, start, end):
        """
        Sets the axis1 limits
        """
        if start > end:
            raise ValueError("Start point over end for this view.")

        self.axis1_limits = start, end

    def set_axis2_limits(self, start, end):
        """
        Sets the axis2 limits
        """
        if start > end:
            raise ValueError("Start point over end for this view.")

        self.axis2_limits = start, end

    def clear_limits(self):
        """
        Clears all limits
        """
        self.axis1_limits = None
        self.axis2_limits = None


import re

class bcolors:
    """
    Helper class that contains formatting classes and functions
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def is_legal_parameter(name):
    """
    Function that returns true if the given name can be used as a
    parameter in the c programming language.
    """
    # List of C reserved keywords
    reserved_keywords = [
        "auto", "break", "case", "char", "const", "continue", "default", "do",
        "double", "else", "enum", "extern", "float", "for", "goto", "if",
        "inline", "int", "long", "register", "restrict", "return", "short",
        "signed", "sizeof", "static", "struct", "switch", "typedef", "union",
        "unsigned", "void", "volatile", "while", "_Bool", "_Complex", "_Imaginary"
    ]

    # Regular expression pattern for a legal C variable name
    pattern = r'^[a-zA-Z_][a-zA-Z0-9_]*$'

    # Check if the input string matches the pattern and is not a reserved keyword
    return re.match(pattern, name) and name not in reserved_keywords


def is_legal_filename(name):
    """
    Function that returns true if the given name can be used as a
    filename
    """

    if name == "":
        return False

    if " " in name:
        return False

    if "/" in name:
        return False

    if "\\" in name:
        return False

    return True



import yaml
import os

from mcstasscript.data.data import McStasData
import mcstasscript.helper.managed_mcrun as managed_mcrun


def name_search(name, data_list):
    """"
    name_search returns McStasData instance with specific name if it is
    in the given data_list. If no match is found, a search for the data
    filename is performed. If several matches are found, a list of
    McStasData objects are returned.

    The index of certain datasets in the data_list can change if
    additional monitors are added so it is more convenient to access
    the data files using their names.

    Parameters
    ----------
    name : string
        Name of the dataset to be retrieved (component_name)

    data_list : List of McStasData instances
        List of datasets to search
    """
    if type(data_list) is not list:
        raise RuntimeError(
            "name_search function needs list of McStasData as input.")

    if len(data_list) == 0:
        raise RuntimeError("Given data list empty.")

    if not isinstance(data_list[0], McStasData):
        raise RuntimeError(
            "name_search function needs objects of type McStasData as input.")

    # Search by component name
    list_result = []
    for check in data_list:
        if check.name == name:
            list_result.append(check)

    if len(list_result) == 0:
        # Search by filename
        for check in data_list:
            if check.metadata.filename == name:
                list_result.append(check)

    if len(list_result) == 0:
        raise NameError("No dataset with name: \""
                        + name
                        + "\" found.")

    if len(list_result) == 1:
        return list_result[0]
    else:
        return list_result


def name_plot_options(name, data_list, **kwargs):
    """"
    name_plot_options passes keyword arguments to dataset with certain
    name in given data list

    Function for quickly setting plotting options on a certain dataset
    n a larger list of datasets

    Parameters
    ----------
    name : string
        Name of the dataset to be modified (component_name)

    data_list : List of McStasData instances
        List of datasets to search

    kwargs : keyword arguments
        Keyword arguments passed to set_plot_options in
        McStasPlotOptions
    """
    object_to_modify = name_search(name, data_list)
    if type(object_to_modify) is not list:
        object_to_modify.set_plot_options(**kwargs)
    else:
        for data_object in object_to_modify:
            data_object.set_plot_options(**kwargs)

def load_data(foldername):
    """
    Loads data from a McStas data folder including mccode.sim

    Parameters
    ----------
        foldername : string
            Name of the folder from which to load data
    """
    if not os.path.isdir(foldername):
        raise RuntimeError("Could not find specified foldername for"
                           + "load_data:" + str(foldername))

    return managed_mcrun.load_results(foldername)

def load_metadata(data_folder_name):
    """
    Function that loads metadata from a mcstas simulation

    Returns list of metadata objects corresponding to each monitor, all
    information is taken from mccode.sim file.

    Parameters
    ----------

    first argument : str
        path to folder from which metadata should be loaded
    """
    return managed_mcrun.load_metadata(data_folder_name)

def load_monitor(metadata, data_folder_name):
    """
    Function that loads data given metadata and name of data folder

    Loads data for single monitor and returns a McStasData object

    Parameters
    ----------

    metadata : McStasMetaData object
        McStasMetaData object corresponding to the monitor to be loaded

    data_folder_name : str
        path to folder from which metadata should be loaded
    """
    return managed_mcrun.load_monitor(metadata, data_folder_name)


class Configurator:
    """
    Class for setting the configuration file for McStasScript.

    Attributes
    ----------
    configuration_file_name : str
        absolute path of configuration file

    Methods
    -------
    set_mcstas_path(string)
        sets mcstas path

    set_mcrun_path(string)
        sets mcrun path

    set_mcxtrace_path(string)
        sets mcxtrace path

    set_mxrun_path(string)
        sets mxrun path

    set_line_length(int)
        sets maximum line length to given int

    _write_yaml(dict)
        internal method, writes a configuration yaml file with dict content

    _read_yaml()
        internal method, reads a configuration yaml file and returns a dict

    _create_new_config_file()
        internal method, creates default configuration file

    """

    def __init__(self, *args):
        """
        Initialization of configurator, checks that the configuration file
        actually exists, and if it does not, creates a default configuration
        file.

        Parameters
        ----------
        (optional) custom name : str
            Custom name for configuration file for testing purposes
        """

        if len(args) == 1:
            name = args[0]
        else:
            name = "configuration"

        # check configuration file exists
        THIS_DIR = os.path.dirname(os.path.abspath(__file__))
        conf_file = os.path.join(THIS_DIR, "..", name + ".yaml")
        self.configuration_file_name = conf_file
        if not os.path.isfile(self.configuration_file_name):
            # no config file found, write default config file
            self._create_new_config_file()

    def _write_yaml(self, dictionary):
        """
        Writes a dictionary as the new configuration file
        """
        with open(self.configuration_file_name, 'w') as yaml_file:
            yaml.dump(dictionary, yaml_file, default_flow_style=False)

    def _read_yaml(self):
        """
        Reads yaml configuration file
        """
        with open(self.configuration_file_name, 'r') as ymlfile:
            return yaml.safe_load(ymlfile)

    def _create_new_config_file(self):
        """
        Writes a default configuration file to the package root directory
        """

        run = "/Applications/McStas-2.5.app/Contents/Resources/mcstas/2.5/bin/"
        mcstas = "/Applications/McStas-2.5.app/Contents/Resources/mcstas/2.5/"

        mxrun = "/Applications/McXtrace-1.5.app" \
                + "/Contents/Resources/mcxtrace/1.5/mxrun"
        mcxtrace = "/Applications/McXtrace-1.5.app" \
                   + "/Contents/Resources/mcxtrace/1.5/"

        default_paths = {"mcrun_path": run,
                         "mcstas_path": mcstas,
                         "mxrun_path": mxrun,
                         "mcxtrace_path": mcxtrace}

        default_other = {"characters_per_line": 85}

        default_config = {"paths": default_paths, "other": default_other}

        self._write_yaml(default_config)

    def set_mcstas_path(self, path):
        """
        Sets the path to McStas

        Parameters
        ----------
        path : str
            Path to the mcstas directory containing "sources", "optics", ...
        """

        if not os.path.isdir(path):
            raise RuntimeError("Invalid path given to set_mcstas_path:"
                               + str(path))

        # read entire configuration file
        config = self._read_yaml()

        # update mcstas_path
        config["paths"]["mcstas_path"] = path

        # write new configuration file
        self._write_yaml(config)

    def set_mcrun_path(self, path):
        """
        Sets the path to mcrun

        Parameters
        ----------
        path : str
            Path to the mcrun executable
        """

        if not os.path.isdir(path):
            raise RuntimeError("Invalid path given to set_mcrun_path:"
                               + str(path))

        # read entire configuration file
        config = self._read_yaml()

        # update mcstas_path
        config["paths"]["mcrun_path"] = path

        # write new configuration file
        self._write_yaml(config)

    def set_mcxtrace_path(self, path):
        """
        Sets the path to McXtrace

        Parameters
        ----------
        path : str
            Path to the mcxtrace directory containing "sources", "optics", ...
        """

        if not os.path.isdir(path):
            raise RuntimeError("Invalid path given to set_mcxtrace_path:"
                               + str(path))

        # read entire configuration file
        config = self._read_yaml()

        # update mcxtrace_path
        config["paths"]["mcxtrace_path"] = path

        # write new configuration file
        self._write_yaml(config)

    def set_mxrun_path(self, path):
        """
        Sets the path to mxrun

        Parameters
        ----------
        path : str
            Path to the mxrun executable
        """

        if not os.path.isdir(path):
            raise RuntimeError("Invalid path given to set_mxrun_path: "
                               + str(path))

        # read entire configuration file
        config = self._read_yaml()

        # update mxrun_path
        config["paths"]["mxrun_path"] = path

        # write new configuration file
        self._write_yaml(config)

    def set_line_length(self, line_length):
        """
        Sets maximum line length for output

        Parameters
        ----------
        line_length : int
            maximum line length for output
        """

        if not isinstance(line_length, int):
            raise ValueError("Given line length in set_line_length not an "
                             + "integer.")

        if line_length < 1:
            raise ValueError("Line length specified in set_line_length must"
                             + " be positve, given length: "
                             + str(line_length))

        # read entire configuration file
        config = self._read_yaml()

        # update mcstas_path
        config["other"]["characters_per_line"] = int(line_length)

        # write new configuration file
        self._write_yaml(config)

    def __repr__(self):
        string = "Configurator:\n"
        config = self._read_yaml()
        if "paths" in config:
            string += " paths:\n"
            for key, value in config["paths"].items():
                string += "  " + str(key) + ": " + str(value) + "\n"

        if "other" in config:
            string += " other:\n"
            for key, value in config["other"].items():
                string += "  " + str(key) + ": " + str(value) + "\n"

        return string



import unittest
import unittest.mock
import io

from mcstasscript.jb_interface.widget_helpers import HiddenPrints
from mcstasscript.jb_interface.widget_helpers import parameter_has_default
from mcstasscript.jb_interface.widget_helpers import get_parameter_default
from mcstasscript.helper.mcstas_objects import provide_parameter


class TestWidgetHelpers(unittest.TestCase):
    """
    Tests of widget helpers
    """

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_HiddenPrints(self, mock_stdout):
        """
        Checks HiddenPrints suppress output to stdout
        """

        print("test")

        output = mock_stdout.getvalue()
        self.assertEqual(output, "test\n")

        with HiddenPrints():
            print("Hello")

        output = mock_stdout.getvalue()
        self.assertEqual(output, "test\n")

    def test_parameter_has_default_false(self):
        """
        Check for parameter that does not have default, should be false
        """

        par = provide_parameter("test")
        self.assertFalse(parameter_has_default(par))

    def test_parameter_has_default_true(self):
        """
        Check for parameter that has default, should be true
        """

        par = provide_parameter("test", value=8)
        self.assertTrue(parameter_has_default(par))

    def test_get_parameter_default_string(self):
        """
        Get the default for string parameter
        """

        par = provide_parameter("string", "test", value="variable")
        self.assertEqual(get_parameter_default(par), "variable")

    def test_get_parameter_default_double_specified(self):
        """
        Get the default for specified double parameter
        """

        par = provide_parameter("double", "test", value=5.5)
        self.assertEqual(get_parameter_default(par), 5.5)

    def test_get_parameter_default_double(self):
        """
        Get the default for parameter that was double by default
        """

        par = provide_parameter("test", value=5.7)
        self.assertEqual(get_parameter_default(par), 5.7)

    def test_get_parameter_default_int(self):
        """
        Get default value from integer value, should be rounded
        """

        par = provide_parameter("int", "test", value=5.7)
        self.assertEqual(get_parameter_default(par), 5)


